from __future__ import unicode_literals
from functools import wraps

import binascii
import datetime
import inspect
import random
import re
import six
import string
from six.moves.urllib.parse import urlparse


REQUEST_ID_LONG = string.digits + string.ascii_uppercase


def camelcase_to_underscores(argument):
    ''' Converts a camelcase param like theNewAttribute to the equivalent
    python underscore variable like the_new_attribute'''
    result = ''
    prev_char_title = True
    if not argument:
        return argument
    for index, char in enumerate(argument):
        try:
            next_char_title = argument[index + 1].istitle()
        except IndexError:
            next_char_title = True

        upper_to_lower = char.istitle() and not next_char_title
        lower_to_upper = char.istitle() and not prev_char_title

        if index and (upper_to_lower or lower_to_upper):
            # Only add underscore if char is capital, not first letter, and next
            # char is not capital
            result += "_"
        prev_char_title = char.istitle()
        if not char.isspace():  # Only add non-whitespace
            result += char.lower()
    return result


def underscores_to_camelcase(argument):
    ''' Converts a camelcase param like the_new_attribute to the equivalent
    camelcase version like theNewAttribute. Note that the first letter is
    NOT capitalized by this function '''
    result = ''
    previous_was_underscore = False
    for char in argument:
        if char != '_':
            if previous_was_underscore:
                result += char.upper()
            else:
                result += char
        previous_was_underscore = char == '_'
    return result


def method_names_from_class(clazz):
    # On Python 2, methods are different from functions, and the `inspect`
    # predicates distinguish between them. On Python 3, methods are just
    # regular functions, and `inspect.ismethod` doesn't work, so we have to
    # use `inspect.isfunction` instead
    if six.PY2:
        predicate = inspect.ismethod
    else:
        predicate = inspect.isfunction
    return [x[0] for x in inspect.getmembers(clazz, predicate=predicate)]


def get_random_hex(length=8):
    chars = list(range(10)) + ['a', 'b', 'c', 'd', 'e', 'f']
    return ''.join(six.text_type(random.choice(chars)) for x in range(length))


def get_random_message_id():
    return '{0}-{1}-{2}-{3}-{4}'.format(get_random_hex(8), get_random_hex(4), get_random_hex(4), get_random_hex(4), get_random_hex(12))


def convert_regex_to_flask_path(url_path):
    """
    Converts a regex matching url to one that can be used with flask
    """
    for token in ["$"]:
        url_path = url_path.replace(token, "")

    def caller(reg):
        match_name, match_pattern = reg.groups()
        return '<regex("{0}"):{1}>'.format(match_pattern, match_name)

    url_path = re.sub("\(\?P<(.*?)>(.*?)\)", caller, url_path)

    if url_path.endswith("/?"):
        # Flask does own handling of trailing slashes
        url_path = url_path.rstrip("/?")
    return url_path


class convert_httpretty_response(object):

    def __init__(self, callback):
        self.callback = callback

    @property
    def __name__(self):
        # For instance methods, use class and method names. Otherwise
        # use module and method name
        if inspect.ismethod(self.callback):
            outer = self.callback.__self__.__class__.__name__
        else:
            outer = self.callback.__module__
        return "{0}.{1}".format(outer, self.callback.__name__)

    def __call__(self, request, url, headers, **kwargs):
        result = self.callback(request, url, headers)
        status, headers, response = result
        if 'server' not in headers:
            headers["server"] = "amazon.com"
        return status, headers, response


class convert_flask_to_httpretty_response(object):

    def __init__(self, callback):
        self.callback = callback

    @property
    def __name__(self):
        # For instance methods, use class and method names. Otherwise
        # use module and method name
        if inspect.ismethod(self.callback):
            outer = self.callback.__self__.__class__.__name__
        else:
            outer = self.callback.__module__
        return "{0}.{1}".format(outer, self.callback.__name__)

    def __call__(self, args=None, **kwargs):
        from flask import request, Response

        result = self.callback(request, request.url, {})
        # result is a status, headers, response tuple
        if len(result) == 3:
            status, headers, content = result
        else:
            status, headers, content = 200, {}, result

        response = Response(response=content, status=status, headers=headers)
        if request.method == "HEAD" and 'content-length' in headers:
            response.headers['Content-Length'] = headers['content-length']
        return response


class convert_flask_to_responses_response(object):

    def __init__(self, callback):
        self.callback = callback

    @property
    def __name__(self):
        # For instance methods, use class and method names. Otherwise
        # use module and method name
        if inspect.ismethod(self.callback):
            outer = self.callback.__self__.__class__.__name__
        else:
            outer = self.callback.__module__
        return "{0}.{1}".format(outer, self.callback.__name__)

    def __call__(self, request, *args, **kwargs):
        for key, val in request.headers.items():
            if isinstance(val, six.binary_type):
                request.headers[key] = val.decode("utf-8")

        result = self.callback(request, request.url, request.headers)
        status, headers, response = result
        return status, headers, response


def iso_8601_datetime_with_milliseconds(datetime):
    return datetime.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + 'Z'


def iso_8601_datetime_without_milliseconds(datetime):
    return datetime.strftime("%Y-%m-%dT%H:%M:%S") + 'Z'


RFC1123 = '%a, %d %b %Y %H:%M:%S GMT'


def rfc_1123_datetime(datetime):
    return datetime.strftime(RFC1123)


def str_to_rfc_1123_datetime(str):
    return datetime.datetime.strptime(str, RFC1123)


def unix_time(dt=None):
    dt = dt or datetime.datetime.utcnow()
    epoch = datetime.datetime.utcfromtimestamp(0)
    delta = dt - epoch
    return (delta.days * 86400) + (delta.seconds + (delta.microseconds / 1e6))


def unix_time_millis(dt=None):
    return unix_time(dt) * 1000.0


def gen_amz_crc32(response, headerdict=None):
    if not isinstance(response, bytes):
        response = response.encode()

    crc = str(binascii.crc32(response))

    if headerdict is not None and isinstance(headerdict, dict):
        headerdict.update({'x-amz-crc32': crc})

    return crc


def gen_amzn_requestid_long(headerdict=None):
    req_id = ''.join([random.choice(REQUEST_ID_LONG) for _ in range(0, 52)])

    if headerdict is not None and isinstance(headerdict, dict):
        headerdict.update({'x-amzn-requestid': req_id})

    return req_id


def amz_crc32(f):
    @wraps(f)
    def _wrapper(*args, **kwargs):
        response = f(*args, **kwargs)

        headers = {}
        status = 200

        if isinstance(response, six.string_types):
            body = response
        else:
            if len(response) == 2:
                body, new_headers = response
                status = new_headers.get('status', 200)
            else:
                status, new_headers, body = response
            headers.update(new_headers)
            # Cast status to string
            if "status" in headers:
                headers['status'] = str(headers['status'])

        try:
            # Doesnt work on python2 for some odd unicode strings
            gen_amz_crc32(body, headers)
        except Exception:
            pass

        return status, headers, body

    return _wrapper


def amzn_request_id(f):
    @wraps(f)
    def _wrapper(*args, **kwargs):
        response = f(*args, **kwargs)

        headers = {}
        status = 200

        if isinstance(response, six.string_types):
            body = response
        else:
            if len(response) == 2:
                body, new_headers = response
                status = new_headers.get('status', 200)
            else:
                status, new_headers, body = response
            headers.update(new_headers)

        request_id = gen_amzn_requestid_long(headers)

        # Update request ID in XML
        try:
            body = re.sub(r'(?<=<RequestId>).*(?=<\/RequestId>)', request_id, body)
        except Exception:  # Will just ignore if it cant work on bytes (which are str's on python2)
            pass

        return status, headers, body

    return _wrapper


def path_url(url):
    parsed_url = urlparse(url)
    path = parsed_url.path
    if not path:
        path = '/'
    if parsed_url.query:
        path = path + '?' + parsed_url.query
    return path



from __future__ import unicode_literals

import datetime
import json

import boto.cognito.identity

from moto.compat import OrderedDict
from moto.core import BaseBackend, BaseModel
from moto.core.utils import iso_8601_datetime_with_milliseconds

from .utils import get_random_identity_id


class CognitoIdentity(BaseModel):

    def __init__(self, region, identity_pool_name, **kwargs):
        self.identity_pool_name = identity_pool_name
        self.allow_unauthenticated_identities = kwargs.get('allow_unauthenticated_identities', '')
        self.supported_login_providers = kwargs.get('supported_login_providers', {})
        self.developer_provider_name = kwargs.get('developer_provider_name', '')
        self.open_id_connect_provider_arns = kwargs.get('open_id_connect_provider_arns', [])
        self.cognito_identity_providers = kwargs.get('cognito_identity_providers', [])
        self.saml_provider_arns = kwargs.get('saml_provider_arns', [])

        self.identity_pool_id = get_random_identity_id(region)
        self.creation_time = datetime.datetime.utcnow()


class CognitoIdentityBackend(BaseBackend):

    def __init__(self, region):
        super(CognitoIdentityBackend, self).__init__()
        self.region = region
        self.identity_pools = OrderedDict()

    def reset(self):
        region = self.region
        self.__dict__ = {}
        self.__init__(region)

    def create_identity_pool(self, identity_pool_name, allow_unauthenticated_identities,
        supported_login_providers, developer_provider_name, open_id_connect_provider_arns,
            cognito_identity_providers, saml_provider_arns):

        new_identity = CognitoIdentity(self.region, identity_pool_name,
            allow_unauthenticated_identities=allow_unauthenticated_identities,
            supported_login_providers=supported_login_providers,
            developer_provider_name=developer_provider_name,
            open_id_connect_provider_arns=open_id_connect_provider_arns,
            cognito_identity_providers=cognito_identity_providers,
            saml_provider_arns=saml_provider_arns)
        self.identity_pools[new_identity.identity_pool_id] = new_identity

        response = json.dumps({
            'IdentityPoolId': new_identity.identity_pool_id,
            'IdentityPoolName': new_identity.identity_pool_name,
            'AllowUnauthenticatedIdentities': new_identity.allow_unauthenticated_identities,
            'SupportedLoginProviders': new_identity.supported_login_providers,
            'DeveloperProviderName': new_identity.developer_provider_name,
            'OpenIdConnectProviderARNs': new_identity.open_id_connect_provider_arns,
            'CognitoIdentityProviders': new_identity.cognito_identity_providers,
            'SamlProviderARNs': new_identity.saml_provider_arns
        })

        return response

    def get_id(self):
        identity_id = {'IdentityId': get_random_identity_id(self.region)}
        return json.dumps(identity_id)

    def get_credentials_for_identity(self, identity_id):
        duration = 90
        now = datetime.datetime.utcnow()
        expiration = now + datetime.timedelta(seconds=duration)
        expiration_str = str(iso_8601_datetime_with_milliseconds(expiration))
        response = json.dumps(
            {
                "Credentials":
                {
                    "AccessKeyId": "TESTACCESSKEY12345",
                    "Expiration": expiration_str,
                    "SecretKey": "ABCSECRETKEY",
                    "SessionToken": "ABC12345"
                },
                "IdentityId": identity_id
            })
        return response

    def get_open_id_token_for_developer_identity(self, identity_id):
        response = json.dumps(
            {
                "IdentityId": identity_id,
                "Token": get_random_identity_id(self.region)
            })
        return response


cognitoidentity_backends = {}
for region in boto.cognito.identity.regions():
    cognitoidentity_backends[region.name] = CognitoIdentityBackend(region.name)



from __future__ import unicode_literals
import boto3
import re
import requests.adapters
from itertools import cycle
import six
import datetime
import time
import uuid
import logging
import docker
import functools
import threading
import dateutil.parser
from moto.core import BaseBackend, BaseModel
from moto.iam import iam_backends
from moto.ec2 import ec2_backends
from moto.ecs import ecs_backends
from moto.logs import logs_backends

from .exceptions import InvalidParameterValueException, InternalFailure, ClientException
from .utils import make_arn_for_compute_env, make_arn_for_job_queue, make_arn_for_task_def, lowercase_first_key
from moto.ec2.exceptions import InvalidSubnetIdError
from moto.ec2.models import INSTANCE_TYPES as EC2_INSTANCE_TYPES
from moto.iam.exceptions import IAMNotFoundException


_orig_adapter_send = requests.adapters.HTTPAdapter.send
logger = logging.getLogger(__name__)
DEFAULT_ACCOUNT_ID = 123456789012
COMPUTE_ENVIRONMENT_NAME_REGEX = re.compile(r'^[A-Za-z0-9][A-Za-z0-9_-]{1,126}[A-Za-z0-9]$')


def datetime2int(date):
    return int(time.mktime(date.timetuple()))


class ComputeEnvironment(BaseModel):
    def __init__(self, compute_environment_name, _type, state, compute_resources, service_role, region_name):
        self.name = compute_environment_name
        self.env_type = _type
        self.state = state
        self.compute_resources = compute_resources
        self.service_role = service_role
        self.arn = make_arn_for_compute_env(DEFAULT_ACCOUNT_ID, compute_environment_name, region_name)

        self.instances = []
        self.ecs_arn = None
        self.ecs_name = None

    def add_instance(self, instance):
        self.instances.append(instance)

    def set_ecs(self, arn, name):
        self.ecs_arn = arn
        self.ecs_name = name

    @property
    def physical_resource_id(self):
        return self.arn

    @classmethod
    def create_from_cloudformation_json(cls, resource_name, cloudformation_json, region_name):
        backend = batch_backends[region_name]
        properties = cloudformation_json['Properties']

        env = backend.create_compute_environment(
            resource_name,
            properties['Type'],
            properties.get('State', 'ENABLED'),
            lowercase_first_key(properties['ComputeResources']),
            properties['ServiceRole']
        )
        arn = env[1]

        return backend.get_compute_environment_by_arn(arn)


class JobQueue(BaseModel):
    def __init__(self, name, priority, state, environments, env_order_json, region_name):
        """
        :param name: Job queue name
        :type name: str
        :param priority: Job queue priority
        :type priority: int
        :param state: Either ENABLED or DISABLED
        :type state: str
        :param environments: Compute Environments
        :type environments: list of ComputeEnvironment
        :param env_order_json: Compute Environments JSON for use when describing
        :type env_order_json: list of dict
        :param region_name: Region name
        :type region_name: str
        """
        self.name = name
        self.priority = priority
        self.state = state
        self.environments = environments
        self.env_order_json = env_order_json
        self.arn = make_arn_for_job_queue(DEFAULT_ACCOUNT_ID, name, region_name)
        self.status = 'VALID'

        self.jobs = []

    def describe(self):
        result = {
            'computeEnvironmentOrder': self.env_order_json,
            'jobQueueArn': self.arn,
            'jobQueueName': self.name,
            'priority': self.priority,
            'state': self.state,
            'status': self.status
        }

        return result

    @property
    def physical_resource_id(self):
        return self.arn

    @classmethod
    def create_from_cloudformation_json(cls, resource_name, cloudformation_json, region_name):
        backend = batch_backends[region_name]
        properties = cloudformation_json['Properties']

        # Need to deal with difference case from cloudformation compute_resources, e.g. instanceRole vs InstanceRole
        # Hacky fix to normalise keys, is making me think I want to start spamming cAsEiNsEnSiTiVe dictionaries
        compute_envs = [lowercase_first_key(dict_item) for dict_item in properties['ComputeEnvironmentOrder']]

        queue = backend.create_job_queue(
            queue_name=resource_name,
            priority=properties['Priority'],
            state=properties.get('State', 'ENABLED'),
            compute_env_order=compute_envs
        )
        arn = queue[1]

        return backend.get_job_queue_by_arn(arn)


class JobDefinition(BaseModel):
    def __init__(self, name, parameters, _type, container_properties, region_name, revision=0, retry_strategy=0):
        self.name = name
        self.retries = retry_strategy
        self.type = _type
        self.revision = revision
        self._region = region_name
        self.container_properties = container_properties
        self.arn = None
        self.status = 'INACTIVE'

        if parameters is None:
            parameters = {}
        self.parameters = parameters

        self._validate()
        self._update_arn()

    def _update_arn(self):
        self.revision += 1
        self.arn = make_arn_for_task_def(DEFAULT_ACCOUNT_ID, self.name, self.revision, self._region)

    def _validate(self):
        if self.type not in ('container',):
            raise ClientException('type must be one of "container"')

        # For future use when containers arnt the only thing in batch
        if self.type != 'container':
            raise NotImplementedError()

        if not isinstance(self.parameters, dict):
            raise ClientException('parameters must be a string to string map')

        if 'image' not in self.container_properties:
            raise ClientException('containerProperties must contain image')

        if 'memory' not in self.container_properties:
            raise ClientException('containerProperties must contain memory')
        if self.container_properties['memory'] < 4:
            raise ClientException('container memory limit must be greater than 4')

        if 'vcpus' not in self.container_properties:
            raise ClientException('containerProperties must contain vcpus')
        if self.container_properties['vcpus'] < 1:
            raise ClientException('container vcpus limit must be greater than 0')

    def update(self, parameters, _type, container_properties, retry_strategy):
        if parameters is None:
            parameters = self.parameters

        if _type is None:
            _type = self.type

        if container_properties is None:
            container_properties = self.container_properties

        if retry_strategy is None:
            retry_strategy = self.retries

        return JobDefinition(self.name, parameters, _type, container_properties, region_name=self._region, revision=self.revision, retry_strategy=retry_strategy)

    def describe(self):
        result = {
            'jobDefinitionArn': self.arn,
            'jobDefinitionName': self.name,
            'parameters': self.parameters,
            'revision': self.revision,
            'status': self.status,
            'type': self.type
        }
        if self.container_properties is not None:
            result['containerProperties'] = self.container_properties
        if self.retries is not None and self.retries > 0:
            result['retryStrategy'] = {'attempts': self.retries}

        return result

    @property
    def physical_resource_id(self):
        return self.arn

    @classmethod
    def create_from_cloudformation_json(cls, resource_name, cloudformation_json, region_name):
        backend = batch_backends[region_name]
        properties = cloudformation_json['Properties']

        res = backend.register_job_definition(
            def_name=resource_name,
            parameters=lowercase_first_key(properties.get('Parameters', {})),
            _type='container',
            retry_strategy=lowercase_first_key(properties['RetryStrategy']),
            container_properties=lowercase_first_key(properties['ContainerProperties'])
        )

        arn = res[1]

        return backend.get_job_definition_by_arn(arn)


class Job(threading.Thread, BaseModel):
    def __init__(self, name, job_def, job_queue, log_backend):
        """
        Docker Job

        :param name: Job Name
        :param job_def: Job definition
        :type: job_def: JobDefinition
        :param job_queue: Job Queue
        :param log_backend: Log backend
        :type log_backend: moto.logs.models.LogsBackend
        """
        threading.Thread.__init__(self)

        self.job_name = name
        self.job_id = str(uuid.uuid4())
        self.job_definition = job_def
        self.job_queue = job_queue
        self.job_state = 'SUBMITTED'  # One of SUBMITTED | PENDING | RUNNABLE | STARTING | RUNNING | SUCCEEDED | FAILED
        self.job_queue.jobs.append(self)
        self.job_started_at = datetime.datetime(1970, 1, 1)
        self.job_stopped_at = datetime.datetime(1970, 1, 1)
        self.job_stopped = False
        self.job_stopped_reason = None

        self.stop = False

        self.daemon = True
        self.name = 'MOTO-BATCH-' + self.job_id

        self.docker_client = docker.from_env()
        self._log_backend = log_backend

        # Unfortunately mocking replaces this method w/o fallback enabled, so we
        # need to replace it if we detect it's been mocked
        if requests.adapters.HTTPAdapter.send != _orig_adapter_send:
            _orig_get_adapter = self.docker_client.api.get_adapter

            def replace_adapter_send(*args, **kwargs):
                adapter = _orig_get_adapter(*args, **kwargs)

                if isinstance(adapter, requests.adapters.HTTPAdapter):
                    adapter.send = functools.partial(_orig_adapter_send, adapter)
                return adapter
            self.docker_client.api.get_adapter = replace_adapter_send

    def describe(self):
        result = {
            'jobDefinition': self.job_definition.arn,
            'jobId': self.job_id,
            'jobName': self.job_name,
            'jobQueue': self.job_queue.arn,
            'startedAt': datetime2int(self.job_started_at),
            'status': self.job_state,
            'dependsOn': []
        }
        if self.job_stopped:
            result['stoppedAt'] = datetime2int(self.job_stopped_at)
            result['container'] = {}
            result['container']['command'] = ['/bin/sh -c "for a in `seq 1 10`; do echo Hello World; sleep 1; done"']
            result['container']['privileged'] = False
            result['container']['readonlyRootFilesystem'] = False
            result['container']['ulimits'] = {}
            result['container']['vcpus'] = 1
            result['container']['volumes'] = ''
            result['container']['logStreamName'] = self.log_stream_name
        if self.job_stopped_reason is not None:
            result['statusReason'] = self.job_stopped_reason
        return result

    def run(self):
        """
        Run the container.

        Logic is as follows:
        Generate container info (eventually from task definition)
        Start container
        Loop whilst not asked to stop and the container is running.
          Get all logs from container between the last time I checked and now.
        Convert logs into cloudwatch format
        Put logs into cloudwatch

        :return:
        """
        try:
            self.job_state = 'PENDING'
            time.sleep(1)

            image = 'alpine:latest'
            cmd = '/bin/sh -c "for a in `seq 1 10`; do echo Hello World; sleep 1; done"'
            name = '{0}-{1}'.format(self.job_name, self.job_id)

            self.job_state = 'RUNNABLE'
            # TODO setup ecs container instance
            time.sleep(1)

            self.job_state = 'STARTING'
            container = self.docker_client.containers.run(
                image, cmd,
                detach=True,
                name=name
            )
            self.job_state = 'RUNNING'
            self.job_started_at = datetime.datetime.now()
            try:
                # Log collection
                logs_stdout = []
                logs_stderr = []
                container.reload()

                # Dodgy hack, we can only check docker logs once a second, but we want to loop more
                # so we can stop if asked to in a quick manner, should all go away if we go async
                # There also be some dodgyness when sending an integer to docker logs and some
                # events seem to be duplicated.
                now = datetime.datetime.now()
                i = 1
                while container.status == 'running' and not self.stop:
                    time.sleep(0.15)
                    if i % 10 == 0:
                        logs_stderr.extend(container.logs(stdout=False, stderr=True, timestamps=True, since=datetime2int(now)).decode().split('\n'))
                        logs_stdout.extend(container.logs(stdout=True, stderr=False, timestamps=True, since=datetime2int(now)).decode().split('\n'))
                        now = datetime.datetime.now()
                        container.reload()
                    i += 1

                # Container should be stopped by this point... unless asked to stop
                if container.status == 'running':
                    container.kill()

                self.job_stopped_at = datetime.datetime.now()
                # Get final logs
                logs_stderr.extend(container.logs(stdout=False, stderr=True, timestamps=True, since=datetime2int(now)).decode().split('\n'))
                logs_stdout.extend(container.logs(stdout=True, stderr=False, timestamps=True, since=datetime2int(now)).decode().split('\n'))

                self.job_state = 'SUCCEEDED' if not self.stop else 'FAILED'

                # Process logs
                logs_stdout = [x for x in logs_stdout if len(x) > 0]
                logs_stderr = [x for x in logs_stderr if len(x) > 0]
                logs = []
                for line in logs_stdout + logs_stderr:
                    date, line = line.split(' ', 1)
                    date = dateutil.parser.parse(date)
                    date = int(date.timestamp())
                    logs.append({'timestamp': date, 'message': line.strip()})

                # Send to cloudwatch
                log_group = '/aws/batch/job'
                stream_name = '{0}/default/{1}'.format(self.job_definition.name, self.job_id)
                self.log_stream_name = stream_name
                self._log_backend.ensure_log_group(log_group, None)
                self._log_backend.create_log_stream(log_group, stream_name)
                self._log_backend.put_log_events(log_group, stream_name, logs, None)

            except Exception as err:
                logger.error('Failed to run AWS Batch container {0}. Error {1}'.format(self.name, err))
                self.job_state = 'FAILED'
                container.kill()
            finally:
                container.remove()
        except Exception as err:
            logger.error('Failed to run AWS Batch container {0}. Error {1}'.format(self.name, err))
            self.job_state = 'FAILED'

        self.job_stopped = True
        self.job_stopped_at = datetime.datetime.now()

    def terminate(self, reason):
        if not self.stop:
            self.stop = True
            self.job_stopped_reason = reason


class BatchBackend(BaseBackend):
    def __init__(self, region_name=None):
        super(BatchBackend, self).__init__()
        self.region_name = region_name

        self._compute_environments = {}
        self._job_queues = {}
        self._job_definitions = {}
        self._jobs = {}

    @property
    def iam_backend(self):
        """
        :return: IAM Backend
        :rtype: moto.iam.models.IAMBackend
        """
        return iam_backends['global']

    @property
    def ec2_backend(self):
        """
        :return: EC2 Backend
        :rtype: moto.ec2.models.EC2Backend
        """
        return ec2_backends[self.region_name]

    @property
    def ecs_backend(self):
        """
        :return: ECS Backend
        :rtype: moto.ecs.models.EC2ContainerServiceBackend
        """
        return ecs_backends[self.region_name]

    @property
    def logs_backend(self):
        """
        :return: ECS Backend
        :rtype: moto.logs.models.LogsBackend
        """
        return logs_backends[self.region_name]

    def reset(self):
        region_name = self.region_name

        for job in self._jobs.values():
            if job.job_state not in ('FAILED', 'SUCCEEDED'):
                job.stop = True
                # Try to join
                job.join(0.2)

        self.__dict__ = {}
        self.__init__(region_name)

    def get_compute_environment_by_arn(self, arn):
        return self._compute_environments.get(arn)

    def get_compute_environment_by_name(self, name):
        for comp_env in self._compute_environments.values():
            if comp_env.name == name:
                return comp_env
        return None

    def get_compute_environment(self, identifier):
        """
        Get compute environment by name or ARN
        :param identifier: Name or ARN
        :type identifier: str

        :return: Compute Environment or None
        :rtype: ComputeEnvironment or None
        """
        env = self.get_compute_environment_by_arn(identifier)
        if env is None:
            env = self.get_compute_environment_by_name(identifier)
        return env

    def get_job_queue_by_arn(self, arn):
        return self._job_queues.get(arn)

    def get_job_queue_by_name(self, name):
        for comp_env in self._job_queues.values():
            if comp_env.name == name:
                return comp_env
        return None

    def get_job_queue(self, identifier):
        """
        Get job queue by name or ARN
        :param identifier: Name or ARN
        :type identifier: str

        :return: Job Queue or None
        :rtype: JobQueue or None
        """
        env = self.get_job_queue_by_arn(identifier)
        if env is None:
            env = self.get_job_queue_by_name(identifier)
        return env

    def get_job_definition_by_arn(self, arn):
        return self._job_definitions.get(arn)

    def get_job_definition_by_name(self, name):
        for comp_env in self._job_definitions.values():
            if comp_env.name == name:
                return comp_env
        return None

    def get_job_definition_by_name_revision(self, name, revision):
        for job_def in self._job_definitions.values():
            if job_def.name == name and job_def.revision == revision:
                return job_def
        return None

    def get_job_definition(self, identifier):
        """
        Get job defintiion by name or ARN
        :param identifier: Name or ARN
        :type identifier: str

        :return: Job definition or None
        :rtype: JobDefinition or None
        """
        env = self.get_job_definition_by_arn(identifier)
        if env is None:
            env = self.get_job_definition_by_name(identifier)
        return env

    def get_job_definitions(self, identifier):
        """
        Get job defintiion by name or ARN
        :param identifier: Name or ARN
        :type identifier: str

        :return: Job definition or None
        :rtype: list of JobDefinition
        """
        result = []
        env = self.get_job_definition_by_arn(identifier)
        if env is not None:
            result.append(env)
        else:
            for value in self._job_definitions.values():
                if value.name == identifier:
                    result.append(value)

        return result

    def get_job_by_id(self, identifier):
        """
        Get job by id
        :param identifier: Job ID
        :type identifier: str

        :return: Job
        :rtype: Job
        """
        try:
            return self._jobs[identifier]
        except KeyError:
            return None

    def describe_compute_environments(self, environments=None, max_results=None, next_token=None):
        envs = set()
        if environments is not None:
            envs = set(environments)

        result = []
        for arn, environment in self._compute_environments.items():
            # Filter shortcut
            if len(envs) > 0 and arn not in envs and environment.name not in envs:
                continue

            json_part = {
                'computeEnvironmentArn': arn,
                'computeEnvironmentName': environment.name,
                'ecsClusterArn': environment.ecs_arn,
                'serviceRole': environment.service_role,
                'state': environment.state,
                'type': environment.env_type,
                'status': 'VALID'
            }
            if environment.env_type == 'MANAGED':
                json_part['computeResources'] = environment.compute_resources

            result.append(json_part)

        return result

    def create_compute_environment(self, compute_environment_name, _type, state, compute_resources, service_role):
        # Validate
        if COMPUTE_ENVIRONMENT_NAME_REGEX.match(compute_environment_name) is None:
            raise InvalidParameterValueException('Compute environment name does not match ^[A-Za-z0-9][A-Za-z0-9_-]{1,126}[A-Za-z0-9]$')

        if self.get_compute_environment_by_name(compute_environment_name) is not None:
            raise InvalidParameterValueException('A compute environment already exists with the name {0}'.format(compute_environment_name))

        # Look for IAM role
        try:
            self.iam_backend.get_role_by_arn(service_role)
        except IAMNotFoundException:
            raise InvalidParameterValueException('Could not find IAM role {0}'.format(service_role))

        if _type not in ('MANAGED', 'UNMANAGED'):
            raise InvalidParameterValueException('type {0} must be one of MANAGED | UNMANAGED'.format(service_role))

        if state is not None and state not in ('ENABLED', 'DISABLED'):
            raise InvalidParameterValueException('state {0} must be one of ENABLED | DISABLED'.format(state))

        if compute_resources is None and _type == 'MANAGED':
            raise InvalidParameterValueException('computeResources must be specified when creating a MANAGED environment'.format(state))
        elif compute_resources is not None:
            self._validate_compute_resources(compute_resources)

        # By here, all values except SPOT ones have been validated
        new_comp_env = ComputeEnvironment(
            compute_environment_name, _type, state,
            compute_resources, service_role,
            region_name=self.region_name
        )
        self._compute_environments[new_comp_env.arn] = new_comp_env

        # Ok by this point, everything is legit, so if its Managed then start some instances
        if _type == 'MANAGED':
            cpus = int(compute_resources.get('desiredvCpus', compute_resources['minvCpus']))
            instance_types = compute_resources['instanceTypes']
            needed_instance_types = self.find_min_instances_to_meet_vcpus(instance_types, cpus)
            # Create instances

            # Will loop over and over so we get decent subnet coverage
            subnet_cycle = cycle(compute_resources['subnets'])

            for instance_type in needed_instance_types:
                reservation = self.ec2_backend.add_instances(
                    image_id='ami-ecs-optimised',  # Todo import AMIs
                    count=1,
                    user_data=None,
                    security_group_names=[],
                    instance_type=instance_type,
                    region_name=self.region_name,
                    subnet_id=six.next(subnet_cycle),
                    key_name=compute_resources.get('ec2KeyPair', 'AWS_OWNED'),
                    security_group_ids=compute_resources['securityGroupIds']
                )

                new_comp_env.add_instance(reservation.instances[0])

        # Create ECS cluster
        # Should be of format P2OnDemand_Batch_UUID
        cluster_name = 'OnDemand_Batch_' + str(uuid.uuid4())
        ecs_cluster = self.ecs_backend.create_cluster(cluster_name)
        new_comp_env.set_ecs(ecs_cluster.arn, cluster_name)

        return compute_environment_name, new_comp_env.arn

    def _validate_compute_resources(self, cr):
        """
        Checks contents of sub dictionary for managed clusters

        :param cr: computeResources
        :type cr: dict
        """
        for param in ('instanceRole', 'maxvCpus', 'minvCpus', 'instanceTypes', 'securityGroupIds', 'subnets', 'type'):
            if param not in cr:
                raise InvalidParameterValueException('computeResources must contain {0}'.format(param))

        if self.iam_backend.get_role_by_arn(cr['instanceRole']) is None:
            raise InvalidParameterValueException('could not find instanceRole {0}'.format(cr['instanceRole']))

        if cr['maxvCpus'] < 0:
            raise InvalidParameterValueException('maxVCpus must be positive')
        if cr['minvCpus'] < 0:
            raise InvalidParameterValueException('minVCpus must be positive')
        if cr['maxvCpus'] < cr['minvCpus']:
            raise InvalidParameterValueException('maxVCpus must be greater than minvCpus')

        if len(cr['instanceTypes']) == 0:
            raise InvalidParameterValueException('At least 1 instance type must be provided')
        for instance_type in cr['instanceTypes']:
            if instance_type == 'optimal':
                pass  # Optimal should pick from latest of current gen
            elif instance_type not in EC2_INSTANCE_TYPES:
                raise InvalidParameterValueException('Instance type {0} does not exist'.format(instance_type))

        for sec_id in cr['securityGroupIds']:
            if self.ec2_backend.get_security_group_from_id(sec_id) is None:
                raise InvalidParameterValueException('security group {0} does not exist'.format(sec_id))
        if len(cr['securityGroupIds']) == 0:
            raise InvalidParameterValueException('At least 1 security group must be provided')

        for subnet_id in cr['subnets']:
            try:
                self.ec2_backend.get_subnet(subnet_id)
            except InvalidSubnetIdError:
                raise InvalidParameterValueException('subnet {0} does not exist'.format(subnet_id))
        if len(cr['subnets']) == 0:
            raise InvalidParameterValueException('At least 1 subnet must be provided')

        if cr['type'] not in ('EC2', 'SPOT'):
            raise InvalidParameterValueException('computeResources.type must be either EC2 | SPOT')

        if cr['type'] == 'SPOT':
            raise InternalFailure('SPOT NOT SUPPORTED YET')

    @staticmethod
    def find_min_instances_to_meet_vcpus(instance_types, target):
        """
        Finds the minimum needed instances to meed a vcpu target

        :param instance_types: Instance types, like ['t2.medium', 't2.small']
        :type instance_types: list of str
        :param target: VCPU target
        :type target: float
        :return: List of instance types
        :rtype: list of str
        """
        # vcpus = [ (vcpus, instance_type), (vcpus, instance_type), ... ]
        instance_vcpus = []
        instances = []

        for instance_type in instance_types:
            if instance_type == 'optimal':
                instance_type = 'm4.4xlarge'

            instance_vcpus.append(
                (EC2_INSTANCE_TYPES[instance_type]['vcpus'], instance_type)
            )

        instance_vcpus = sorted(instance_vcpus, key=lambda item: item[0], reverse=True)
        # Loop through,
        #   if biggest instance type smaller than target, and len(instance_types)> 1, then use biggest type
        #   if biggest instance type bigger than target, and len(instance_types)> 1, then remove it and move on

        #   if biggest instance type bigger than target and len(instan_types) == 1 then add instance and finish
        #   if biggest instance type smaller than target and len(instan_types) == 1 then loop adding instances until target == 0
        #   ^^ boils down to keep adding last till target vcpus is negative
        #   #Algorithm ;-) ... Could probably be done better with some quality lambdas
        while target > 0:
            current_vcpu, current_instance = instance_vcpus[0]

            if len(instance_vcpus) > 1:
                if current_vcpu <= target:
                    target -= current_vcpu
                    instances.append(current_instance)
                else:
                    # try next biggest instance
                    instance_vcpus.pop(0)
            else:
                # Were on the last instance
                target -= current_vcpu
                instances.append(current_instance)

        return instances

    def delete_compute_environment(self, compute_environment_name):
        if compute_environment_name is None:
            raise InvalidParameterValueException('Missing computeEnvironment parameter')

        compute_env = self.get_compute_environment(compute_environment_name)

        if compute_env is not None:
            # Pop ComputeEnvironment
            self._compute_environments.pop(compute_env.arn)

            # Delete ECS cluster
            self.ecs_backend.delete_cluster(compute_env.ecs_name)

            if compute_env.env_type == 'MANAGED':
                # Delete compute envrionment
                instance_ids = [instance.id for instance in compute_env.instances]
                self.ec2_backend.terminate_instances(instance_ids)

    def update_compute_environment(self, compute_environment_name, state, compute_resources, service_role):
        # Validate
        compute_env = self.get_compute_environment(compute_environment_name)
        if compute_env is None:
            raise ClientException('Compute environment {0} does not exist')

        # Look for IAM role
        if service_role is not None:
            try:
                role = self.iam_backend.get_role_by_arn(service_role)
            except IAMNotFoundException:
                raise InvalidParameterValueException('Could not find IAM role {0}'.format(service_role))

            compute_env.service_role = role

        if state is not None:
            if state not in ('ENABLED', 'DISABLED'):
                raise InvalidParameterValueException('state {0} must be one of ENABLED | DISABLED'.format(state))

            compute_env.state = state

        if compute_resources is not None:
            # TODO Implement resizing of instances based on changing vCpus
            # compute_resources CAN contain desiredvCpus, maxvCpus, minvCpus, and can contain none of them.
            pass

        return compute_env.name, compute_env.arn

    def create_job_queue(self, queue_name, priority, state, compute_env_order):
        """
        Create a job queue

        :param queue_name: Queue name
        :type queue_name: str
        :param priority: Queue priority
        :type priority: int
        :param state: Queue state
        :type state: string
        :param compute_env_order: Compute environment list
        :type compute_env_order: list of dict
        :return: Tuple of Name, ARN
        :rtype: tuple of str
        """
        for variable, var_name in ((queue_name, 'jobQueueName'), (priority, 'priority'), (state, 'state'), (compute_env_order, 'computeEnvironmentOrder')):
            if variable is None:
                raise ClientException('{0} must be provided'.format(var_name))

        if state not in ('ENABLED', 'DISABLED'):
            raise ClientException('state {0} must be one of ENABLED | DISABLED'.format(state))
        if self.get_job_queue_by_name(queue_name) is not None:
            raise ClientException('Job queue {0} already exists'.format(queue_name))

        if len(compute_env_order) == 0:
            raise ClientException('At least 1 compute environment must be provided')
        try:
            # orders and extracts computeEnvironment names
            ordered_compute_environments = [item['computeEnvironment'] for item in sorted(compute_env_order, key=lambda x: x['order'])]
            env_objects = []
            # Check each ARN exists, then make a list of compute env's
            for arn in ordered_compute_environments:
                env = self.get_compute_environment_by_arn(arn)
                if env is None:
                    raise ClientException('Compute environment {0} does not exist'.format(arn))
                env_objects.append(env)
        except Exception:
            raise ClientException('computeEnvironmentOrder is malformed')

        # Create new Job Queue
        queue = JobQueue(queue_name, priority, state, env_objects, compute_env_order, self.region_name)
        self._job_queues[queue.arn] = queue

        return queue_name, queue.arn

    def describe_job_queues(self, job_queues=None, max_results=None, next_token=None):
        envs = set()
        if job_queues is not None:
            envs = set(job_queues)

        result = []
        for arn, job_queue in self._job_queues.items():
            # Filter shortcut
            if len(envs) > 0 and arn not in envs and job_queue.name not in envs:
                continue

            result.append(job_queue.describe())

        return result

    def update_job_queue(self, queue_name, priority, state, compute_env_order):
        """
        Update a job queue

        :param queue_name: Queue name
        :type queue_name: str
        :param priority: Queue priority
        :type priority: int
        :param state: Queue state
        :type state: string
        :param compute_env_order: Compute environment list
        :type compute_env_order: list of dict
        :return: Tuple of Name, ARN
        :rtype: tuple of str
        """
        if queue_name is None:
            raise ClientException('jobQueueName must be provided')

        job_queue = self.get_job_queue(queue_name)
        if job_queue is None:
            raise ClientException('Job queue {0} does not exist'.format(queue_name))

        if state is not None:
            if state not in ('ENABLED', 'DISABLED'):
                raise ClientException('state {0} must be one of ENABLED | DISABLED'.format(state))

            job_queue.state = state

        if compute_env_order is not None:
            if len(compute_env_order) == 0:
                raise ClientException('At least 1 compute environment must be provided')
            try:
                # orders and extracts computeEnvironment names
                ordered_compute_environments = [item['computeEnvironment'] for item in sorted(compute_env_order, key=lambda x: x['order'])]
                env_objects = []
                # Check each ARN exists, then make a list of compute env's
                for arn in ordered_compute_environments:
                    env = self.get_compute_environment_by_arn(arn)
                    if env is None:
                        raise ClientException('Compute environment {0} does not exist'.format(arn))
                    env_objects.append(env)
            except Exception:
                raise ClientException('computeEnvironmentOrder is malformed')

            job_queue.env_order_json = compute_env_order
            job_queue.environments = env_objects

        if priority is not None:
            job_queue.priority = priority

        return queue_name, job_queue.arn

    def delete_job_queue(self, queue_name):
        job_queue = self.get_job_queue(queue_name)

        if job_queue is not None:
            del self._job_queues[job_queue.arn]

    def register_job_definition(self, def_name, parameters, _type, retry_strategy, container_properties):
        if def_name is None:
            raise ClientException('jobDefinitionName must be provided')

        job_def = self.get_job_definition_by_name(def_name)
        if retry_strategy is not None:
            try:
                retry_strategy = retry_strategy['attempts']
            except Exception:
                raise ClientException('retryStrategy is malformed')

        if job_def is None:
            job_def = JobDefinition(def_name, parameters, _type, container_properties, region_name=self.region_name, retry_strategy=retry_strategy)
        else:
            # Make new jobdef
            job_def = job_def.update(parameters, _type, container_properties, retry_strategy)

        self._job_definitions[job_def.arn] = job_def

        return def_name, job_def.arn, job_def.revision

    def deregister_job_definition(self, def_name):
        job_def = self.get_job_definition_by_arn(def_name)
        if job_def is None and ':' in def_name:
            name, revision = def_name.split(':', 1)
            job_def = self.get_job_definition_by_name_revision(name, revision)

        if job_def is not None:
            del self._job_definitions[job_def.arn]

    def describe_job_definitions(self, job_def_name=None, job_def_list=None, status=None, max_results=None, next_token=None):
        jobs = []

        # As a job name can reference multiple revisions, we get a list of them
        if job_def_name is not None:
            job_def = self.get_job_definitions(job_def_name)
            if job_def is not None:
                jobs.extend(job_def)
        elif job_def_list is not None:
            for job in job_def_list:
                job_def = self.get_job_definitions(job)
                if job_def is not None:
                    jobs.extend(job_def)
        else:
            jobs.extend(self._job_definitions.values())

        # Got all the job defs were after, filter then by status
        if status is not None:
            return [job for job in jobs if job.status == status]
        return jobs

    def submit_job(self, job_name, job_def_id, job_queue, parameters=None, retries=None, depends_on=None, container_overrides=None):
        # TODO parameters, retries (which is a dict raw from request), job dependancies and container overrides are ignored for now

        # Look for job definition
        job_def = self.get_job_definition_by_arn(job_def_id)
        if job_def is None and ':' in job_def_id:
            job_def = self.get_job_definition_by_name_revision(*job_def_id.split(':', 1))
        if job_def is None:
            raise ClientException('Job definition {0} does not exist'.format(job_def_id))

        queue = self.get_job_queue(job_queue)
        if queue is None:
            raise ClientException('Job queue {0} does not exist'.format(job_queue))

        job = Job(job_name, job_def, queue, log_backend=self.logs_backend)
        self._jobs[job.job_id] = job

        # Here comes the fun
        job.start()

        return job_name, job.job_id

    def describe_jobs(self, jobs):
        job_filter = set()
        if jobs is not None:
            job_filter = set(jobs)

        result = []
        for key, job in self._jobs.items():
            if len(job_filter) > 0 and key not in job_filter:
                continue

            result.append(job.describe())

        return result

    def list_jobs(self, job_queue, job_status=None, max_results=None, next_token=None):
        jobs = []

        job_queue = self.get_job_queue(job_queue)
        if job_queue is None:
            raise ClientException('Job queue {0} does not exist'.format(job_queue))

        if job_status is not None and job_status not in ('SUBMITTED', 'PENDING', 'RUNNABLE', 'STARTING', 'RUNNING', 'SUCCEEDED', 'FAILED'):
            raise ClientException('Job status is not one of SUBMITTED | PENDING | RUNNABLE | STARTING | RUNNING | SUCCEEDED | FAILED')

        for job in job_queue.jobs:
            if job_status is not None and job.job_state != job_status:
                continue

            jobs.append(job)

        return jobs

    def terminate_job(self, job_id, reason):
        if job_id is None:
            raise ClientException('Job ID does not exist')
        if reason is None:
            raise ClientException('Reason does not exist')

        job = self.get_job_by_id(job_id)
        if job is None:
            raise ClientException('Job not found')

        job.terminate(reason)


available_regions = boto3.session.Session().get_available_regions("batch")
batch_backends = {region: BatchBackend(region_name=region) for region in available_regions}



def make_arn_for_load_balancer(account_id, name, region_name):
    return "arn:aws:elasticloadbalancing:{}:{}:loadbalancer/{}/50dc6c495c0c9188".format(
        region_name, account_id, name)


def make_arn_for_target_group(account_id, name, region_name):
    return "arn:aws:elasticloadbalancing:{}:{}:targetgroup/{}/50dc6c495c0c9188".format(
        region_name, account_id, name)



from __future__ import unicode_literals

import json
from werkzeug.exceptions import BadRequest


class RDSClientError(BadRequest):

    def __init__(self, code, message):
        super(RDSClientError, self).__init__()
        self.description = json.dumps({
            "Error": {
                "Code": code,
                "Message": message,
                'Type': 'Sender',
            },
            'RequestId': '6876f774-7273-11e4-85dc-39e55ca848d1',
        })


class DBInstanceNotFoundError(RDSClientError):

    def __init__(self, database_identifier):
        super(DBInstanceNotFoundError, self).__init__(
            'DBInstanceNotFound',
            "Database {0} not found.".format(database_identifier))


class DBSecurityGroupNotFoundError(RDSClientError):

    def __init__(self, security_group_name):
        super(DBSecurityGroupNotFoundError, self).__init__(
            'DBSecurityGroupNotFound',
            "Security Group {0} not found.".format(security_group_name))


class DBSubnetGroupNotFoundError(RDSClientError):

    def __init__(self, subnet_group_name):
        super(DBSubnetGroupNotFoundError, self).__init__(
            'DBSubnetGroupNotFound',
            "Subnet Group {0} not found.".format(subnet_group_name))



from __future__ import unicode_literals

import boto3
import json
from botocore.exceptions import ClientError
from six.moves.email_mime_multipart import MIMEMultipart
from six.moves.email_mime_text import MIMEText

import sure  # noqa
from nose import tools
from moto import mock_ses, mock_sns, mock_sqs
from moto.ses.models import SESFeedback


@mock_ses
def test_enable_disable_ses_sns_communication():
    conn = boto3.client('ses', region_name='us-east-1')
    conn.set_identity_notification_topic(
        Identity='test.com',
        NotificationType='Bounce',
        SnsTopic='the-arn'
    )
    conn.set_identity_notification_topic(
        Identity='test.com',
        NotificationType='Bounce'
    )


def __setup_feedback_env__(ses_conn, sns_conn, sqs_conn, domain, topic, queue, region, expected_msg):
    """Setup the AWS environment to test the SES SNS Feedback"""
    # Environment setup
    # Create SQS queue
    sqs_conn.create_queue(QueueName=queue)
    # Create SNS topic
    create_topic_response = sns_conn.create_topic(Name=topic)
    topic_arn = create_topic_response["TopicArn"]
    # Subscribe the SNS topic to the SQS queue
    sns_conn.subscribe(TopicArn=topic_arn,
                   Protocol="sqs",
                   Endpoint="arn:aws:sqs:%s:123456789012:%s" % (region, queue))
    # Verify SES domain
    ses_conn.verify_domain_identity(Domain=domain)
    # Setup SES notification topic
    if expected_msg is not None:
        ses_conn.set_identity_notification_topic(
            Identity=domain,
            NotificationType=expected_msg,
            SnsTopic=topic_arn
        )


def __test_sns_feedback__(addr, expected_msg):
    region_name = "us-east-1"
    ses_conn = boto3.client('ses', region_name=region_name)
    sns_conn = boto3.client('sns', region_name=region_name)
    sqs_conn = boto3.resource('sqs', region_name=region_name)
    domain = "example.com"
    topic = "bounce-arn-feedback"
    queue = "feedback-test-queue"

    __setup_feedback_env__(ses_conn, sns_conn, sqs_conn, domain, topic, queue, region_name, expected_msg)

    # Send the message
    kwargs = dict(
        Source="test@" + domain,
        Destination={
            "ToAddresses": [addr + "@" + domain],
            "CcAddresses": ["test_cc@" + domain],
            "BccAddresses": ["test_bcc@" + domain],
        },
        Message={
            "Subject": {"Data": "test subject"},
            "Body": {"Text": {"Data": "test body"}}
        }
    )
    ses_conn.send_email(**kwargs)

    # Wait for messages in the queues
    queue = sqs_conn.get_queue_by_name(QueueName=queue)
    messages = queue.receive_messages(MaxNumberOfMessages=1)
    if expected_msg is not None:
        msg = messages[0].body
        msg = json.loads(msg)
        assert msg["Message"] == SESFeedback.generate_message(expected_msg)
    else:
        assert len(messages) == 0


@mock_sqs
@mock_sns
@mock_ses
def test_no_sns_feedback():
    __test_sns_feedback__("test", None)


@mock_sqs
@mock_sns
@mock_ses
def test_sns_feedback_bounce():
    __test_sns_feedback__(SESFeedback.BOUNCE_ADDR, SESFeedback.BOUNCE)


@mock_sqs
@mock_sns
@mock_ses
def test_sns_feedback_complaint():
    __test_sns_feedback__(SESFeedback.COMPLAINT_ADDR, SESFeedback.COMPLAINT)


@mock_sqs
@mock_sns
@mock_ses
def test_sns_feedback_delivery():
    __test_sns_feedback__(SESFeedback.SUCCESS_ADDR, SESFeedback.DELIVERY)



from __future__ import unicode_literals
import collections
import functools
import logging
import copy
import warnings
import re

from moto.autoscaling import models as autoscaling_models
from moto.awslambda import models as lambda_models
from moto.batch import models as batch_models
from moto.cloudwatch import models as cloudwatch_models
from moto.cognitoidentity import models as cognitoidentity_models
from moto.datapipeline import models as datapipeline_models
from moto.dynamodb2 import models as dynamodb2_models
from moto.ec2 import models as ec2_models
from moto.ecs import models as ecs_models
from moto.elb import models as elb_models
from moto.elbv2 import models as elbv2_models
from moto.iam import models as iam_models
from moto.kinesis import models as kinesis_models
from moto.kms import models as kms_models
from moto.rds import models as rds_models
from moto.rds2 import models as rds2_models
from moto.redshift import models as redshift_models
from moto.route53 import models as route53_models
from moto.s3 import models as s3_models
from moto.sns import models as sns_models
from moto.sqs import models as sqs_models
from .utils import random_suffix
from .exceptions import ExportNotFound, MissingParameterError, UnformattedGetAttTemplateException, ValidationError
from boto.cloudformation.stack import Output

MODEL_MAP = {
    "AWS::AutoScaling::AutoScalingGroup": autoscaling_models.FakeAutoScalingGroup,
    "AWS::AutoScaling::LaunchConfiguration": autoscaling_models.FakeLaunchConfiguration,
    "AWS::Batch::JobDefinition": batch_models.JobDefinition,
    "AWS::Batch::JobQueue": batch_models.JobQueue,
    "AWS::Batch::ComputeEnvironment": batch_models.ComputeEnvironment,
    "AWS::DynamoDB::Table": dynamodb2_models.Table,
    "AWS::Kinesis::Stream": kinesis_models.Stream,
    "AWS::Lambda::EventSourceMapping": lambda_models.EventSourceMapping,
    "AWS::Lambda::Function": lambda_models.LambdaFunction,
    "AWS::Lambda::Version": lambda_models.LambdaVersion,
    "AWS::EC2::EIP": ec2_models.ElasticAddress,
    "AWS::EC2::Instance": ec2_models.Instance,
    "AWS::EC2::InternetGateway": ec2_models.InternetGateway,
    "AWS::EC2::NatGateway": ec2_models.NatGateway,
    "AWS::EC2::NetworkInterface": ec2_models.NetworkInterface,
    "AWS::EC2::Route": ec2_models.Route,
    "AWS::EC2::RouteTable": ec2_models.RouteTable,
    "AWS::EC2::SecurityGroup": ec2_models.SecurityGroup,
    "AWS::EC2::SecurityGroupIngress": ec2_models.SecurityGroupIngress,
    "AWS::EC2::SpotFleet": ec2_models.SpotFleetRequest,
    "AWS::EC2::Subnet": ec2_models.Subnet,
    "AWS::EC2::SubnetRouteTableAssociation": ec2_models.SubnetRouteTableAssociation,
    "AWS::EC2::Volume": ec2_models.Volume,
    "AWS::EC2::VolumeAttachment": ec2_models.VolumeAttachment,
    "AWS::EC2::VPC": ec2_models.VPC,
    "AWS::EC2::VPCGatewayAttachment": ec2_models.VPCGatewayAttachment,
    "AWS::EC2::VPCPeeringConnection": ec2_models.VPCPeeringConnection,
    "AWS::ECS::Cluster": ecs_models.Cluster,
    "AWS::ECS::TaskDefinition": ecs_models.TaskDefinition,
    "AWS::ECS::Service": ecs_models.Service,
    "AWS::ElasticLoadBalancing::LoadBalancer": elb_models.FakeLoadBalancer,
    "AWS::ElasticLoadBalancingV2::LoadBalancer": elbv2_models.FakeLoadBalancer,
    "AWS::ElasticLoadBalancingV2::TargetGroup": elbv2_models.FakeTargetGroup,
    "AWS::ElasticLoadBalancingV2::Listener": elbv2_models.FakeListener,
    "AWS::Cognito::IdentityPool": cognitoidentity_models.CognitoIdentity,
    "AWS::DataPipeline::Pipeline": datapipeline_models.Pipeline,
    "AWS::IAM::InstanceProfile": iam_models.InstanceProfile,
    "AWS::IAM::Role": iam_models.Role,
    "AWS::KMS::Key": kms_models.Key,
    "AWS::Logs::LogGroup": cloudwatch_models.LogGroup,
    "AWS::RDS::DBInstance": rds_models.Database,
    "AWS::RDS::DBSecurityGroup": rds_models.SecurityGroup,
    "AWS::RDS::DBSubnetGroup": rds_models.SubnetGroup,
    "AWS::RDS::DBParameterGroup": rds2_models.DBParameterGroup,
    "AWS::Redshift::Cluster": redshift_models.Cluster,
    "AWS::Redshift::ClusterParameterGroup": redshift_models.ParameterGroup,
    "AWS::Redshift::ClusterSubnetGroup": redshift_models.SubnetGroup,
    "AWS::Route53::HealthCheck": route53_models.HealthCheck,
    "AWS::Route53::HostedZone": route53_models.FakeZone,
    "AWS::Route53::RecordSet": route53_models.RecordSet,
    "AWS::Route53::RecordSetGroup": route53_models.RecordSetGroup,
    "AWS::SNS::Topic": sns_models.Topic,
    "AWS::S3::Bucket": s3_models.FakeBucket,
    "AWS::SQS::Queue": sqs_models.Queue,
}

# http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-name.html
NAME_TYPE_MAP = {
    "AWS::CloudWatch::Alarm": "Alarm",
    "AWS::DynamoDB::Table": "TableName",
    "AWS::ElastiCache::CacheCluster": "ClusterName",
    "AWS::ElasticBeanstalk::Application": "ApplicationName",
    "AWS::ElasticBeanstalk::Environment": "EnvironmentName",
    "AWS::ElasticLoadBalancing::LoadBalancer": "LoadBalancerName",
    "AWS::ElasticLoadBalancingV2::TargetGroup": "Name",
    "AWS::RDS::DBInstance": "DBInstanceIdentifier",
    "AWS::S3::Bucket": "BucketName",
    "AWS::SNS::Topic": "TopicName",
    "AWS::SQS::Queue": "QueueName"
}

# Just ignore these models types for now
NULL_MODELS = [
    "AWS::CloudFormation::WaitCondition",
    "AWS::CloudFormation::WaitConditionHandle",
]

DEFAULT_REGION = 'us-east-1'

logger = logging.getLogger("moto")


class LazyDict(dict):

    def __getitem__(self, key):
        val = dict.__getitem__(self, key)
        if callable(val):
            val = val()
            self[key] = val
        return val


def clean_json(resource_json, resources_map):
    """
    Cleanup the a resource dict. For now, this just means replacing any Ref node
    with the corresponding physical_resource_id.

    Eventually, this is where we would add things like function parsing (fn::)
    """
    if isinstance(resource_json, dict):
        if 'Ref' in resource_json:
            # Parse resource reference
            resource = resources_map[resource_json['Ref']]
            if hasattr(resource, 'physical_resource_id'):
                return resource.physical_resource_id
            else:
                return resource

        if "Fn::FindInMap" in resource_json:
            map_name = resource_json["Fn::FindInMap"][0]
            map_path = resource_json["Fn::FindInMap"][1:]
            result = resources_map[map_name]
            for path in map_path:
                result = result[clean_json(path, resources_map)]
            return result

        if 'Fn::GetAtt' in resource_json:
            resource = resources_map.get(resource_json['Fn::GetAtt'][0])
            if resource is None:
                return resource_json
            try:
                return resource.get_cfn_attribute(resource_json['Fn::GetAtt'][1])
            except NotImplementedError as n:
                logger.warning(str(n).format(
                    resource_json['Fn::GetAtt'][0]))
            except UnformattedGetAttTemplateException:
                raise ValidationError(
                    'Bad Request',
                    UnformattedGetAttTemplateException.description.format(
                        resource_json['Fn::GetAtt'][0], resource_json['Fn::GetAtt'][1]))

        if 'Fn::If' in resource_json:
            condition_name, true_value, false_value = resource_json['Fn::If']
            if resources_map.lazy_condition_map[condition_name]:
                return clean_json(true_value, resources_map)
            else:
                return clean_json(false_value, resources_map)

        if 'Fn::Join' in resource_json:
            join_list = clean_json(resource_json['Fn::Join'][1], resources_map)
            return resource_json['Fn::Join'][0].join([str(x) for x in join_list])

        if 'Fn::Split' in resource_json:
            to_split = clean_json(resource_json['Fn::Split'][1], resources_map)
            return to_split.split(resource_json['Fn::Split'][0])

        if 'Fn::Select' in resource_json:
            select_index = int(resource_json['Fn::Select'][0])
            select_list = clean_json(resource_json['Fn::Select'][1], resources_map)
            return select_list[select_index]

        if 'Fn::Sub' in resource_json:
            if isinstance(resource_json['Fn::Sub'], list):
                warnings.warn(
                    "Tried to parse Fn::Sub with variable mapping but it's not supported by moto's CloudFormation implementation")
            else:
                fn_sub_value = clean_json(resource_json['Fn::Sub'], resources_map)
                to_sub = re.findall('(?=\${)[^!^"]*?}', fn_sub_value)
                literals = re.findall('(?=\${!)[^"]*?}', fn_sub_value)
                for sub in to_sub:
                    if '.' in sub:
                        cleaned_ref = clean_json({'Fn::GetAtt': re.findall('(?<=\${)[^"]*?(?=})', sub)[0].split('.')}, resources_map)
                    else:
                        cleaned_ref = clean_json({'Ref': re.findall('(?<=\${)[^"]*?(?=})', sub)[0]}, resources_map)
                    fn_sub_value = fn_sub_value.replace(sub, cleaned_ref)
                for literal in literals:
                    fn_sub_value = fn_sub_value.replace(literal, literal.replace('!', ''))
                return fn_sub_value
            pass

        if 'Fn::ImportValue' in resource_json:
            cleaned_val = clean_json(resource_json['Fn::ImportValue'], resources_map)
            values = [x.value for x in resources_map.cross_stack_resources.values() if x.name == cleaned_val]
            if any(values):
                return values[0]
            else:
                raise ExportNotFound(cleaned_val)

        if 'Fn::GetAZs' in resource_json:
            region = resource_json.get('Fn::GetAZs') or DEFAULT_REGION
            result = []
            # TODO: make this configurable, to reflect the real AWS AZs
            for az in ('a', 'b', 'c', 'd'):
                result.append('%s%s' % (region, az))
            return result

        cleaned_json = {}
        for key, value in resource_json.items():
            cleaned_val = clean_json(value, resources_map)
            if cleaned_val is None:
                # If we didn't find anything, don't add this attribute
                continue
            cleaned_json[key] = cleaned_val
        return cleaned_json
    elif isinstance(resource_json, list):
        return [clean_json(val, resources_map) for val in resource_json]
    else:
        return resource_json


def resource_class_from_type(resource_type):
    if resource_type in NULL_MODELS:
        return None
    if resource_type not in MODEL_MAP:
        logger.warning("No Moto CloudFormation support for %s", resource_type)
        return None
    return MODEL_MAP.get(resource_type)


def resource_name_property_from_type(resource_type):
    return NAME_TYPE_MAP.get(resource_type)


def generate_resource_name(resource_type, stack_name, logical_id):
    if resource_type in ["AWS::ElasticLoadBalancingV2::TargetGroup",
                         "AWS::ElasticLoadBalancingV2::LoadBalancer"]:
        # Target group names need to be less than 32 characters, so when cloudformation creates a name for you
        # it makes sure to stay under that limit
        name_prefix = '{0}-{1}'.format(stack_name, logical_id)
        my_random_suffix = random_suffix()
        truncated_name_prefix = name_prefix[0:32 - (len(my_random_suffix) + 1)]
        # if the truncated name ends in a dash, we'll end up with a double dash in the final name, which is
        # not allowed
        if truncated_name_prefix.endswith('-'):
            truncated_name_prefix = truncated_name_prefix[:-1]
        return '{0}-{1}'.format(truncated_name_prefix, my_random_suffix)
    else:
        return '{0}-{1}-{2}'.format(stack_name, logical_id, random_suffix())


def parse_resource(logical_id, resource_json, resources_map):
    resource_type = resource_json['Type']
    resource_class = resource_class_from_type(resource_type)
    if not resource_class:
        warnings.warn(
            "Tried to parse {0} but it's not supported by moto's CloudFormation implementation".format(resource_type))
        return None

    resource_json = clean_json(resource_json, resources_map)
    resource_name_property = resource_name_property_from_type(resource_type)
    if resource_name_property:
        if 'Properties' not in resource_json:
            resource_json['Properties'] = dict()
        if resource_name_property not in resource_json['Properties']:
            resource_json['Properties'][resource_name_property] = generate_resource_name(
                resource_type, resources_map.get('AWS::StackName'), logical_id)
        resource_name = resource_json['Properties'][resource_name_property]
    else:
        resource_name = generate_resource_name(resource_type, resources_map.get('AWS::StackName'), logical_id)

    return resource_class, resource_json, resource_name


def parse_and_create_resource(logical_id, resource_json, resources_map, region_name):
    condition = resource_json.get('Condition')
    if condition and not resources_map.lazy_condition_map[condition]:
        # If this has a False condition, don't create the resource
        return None

    resource_type = resource_json['Type']
    resource_tuple = parse_resource(logical_id, resource_json, resources_map)
    if not resource_tuple:
        return None
    resource_class, resource_json, resource_name = resource_tuple
    resource = resource_class.create_from_cloudformation_json(
        resource_name, resource_json, region_name)
    resource.type = resource_type
    resource.logical_resource_id = logical_id
    return resource


def parse_and_update_resource(logical_id, resource_json, resources_map, region_name):
    resource_class, new_resource_json, new_resource_name = parse_resource(
        logical_id, resource_json, resources_map)
    original_resource = resources_map[logical_id]
    new_resource = resource_class.update_from_cloudformation_json(
        original_resource=original_resource,
        new_resource_name=new_resource_name,
        cloudformation_json=new_resource_json,
        region_name=region_name
    )
    new_resource.type = resource_json['Type']
    new_resource.logical_resource_id = logical_id
    return new_resource


def parse_and_delete_resource(logical_id, resource_json, resources_map, region_name):
    resource_class, resource_json, resource_name = parse_resource(
        logical_id, resource_json, resources_map)
    resource_class.delete_from_cloudformation_json(
        resource_name, resource_json, region_name)


def parse_condition(condition, resources_map, condition_map):
    if isinstance(condition, bool):
        return condition

    condition_operator = list(condition.keys())[0]

    condition_values = []
    for value in list(condition.values())[0]:
        # Check if we are referencing another Condition
        if 'Condition' in value:
            condition_values.append(condition_map[value['Condition']])
        else:
            condition_values.append(clean_json(value, resources_map))

    if condition_operator == "Fn::Equals":
        return condition_values[0] == condition_values[1]
    elif condition_operator == "Fn::Not":
        return not parse_condition(condition_values[0], resources_map, condition_map)
    elif condition_operator == "Fn::And":
        return all([
            parse_condition(condition_value, resources_map, condition_map)
            for condition_value
            in condition_values])
    elif condition_operator == "Fn::Or":
        return any([
            parse_condition(condition_value, resources_map, condition_map)
            for condition_value
            in condition_values])


def parse_output(output_logical_id, output_json, resources_map):
    output_json = clean_json(output_json, resources_map)
    output = Output()
    output.key = output_logical_id
    output.value = clean_json(output_json['Value'], resources_map)
    output.description = output_json.get('Description')
    return output


class ResourceMap(collections.Mapping):
    """
    This is a lazy loading map for resources. This allows us to create resources
    without needing to create a full dependency tree. Upon creation, each
    each resources is passed this lazy map that it can grab dependencies from.
    """

    def __init__(self, stack_id, stack_name, parameters, tags, region_name, template, cross_stack_resources):
        self._template = template
        self._resource_json_map = template['Resources']
        self._region_name = region_name
        self.input_parameters = parameters
        self.tags = copy.deepcopy(tags)
        self.resolved_parameters = {}
        self.cross_stack_resources = cross_stack_resources

        # Create the default resources
        self._parsed_resources = {
            "AWS::AccountId": "123456789012",
            "AWS::Region": self._region_name,
            "AWS::StackId": stack_id,
            "AWS::StackName": stack_name,
            "AWS::URLSuffix": "amazonaws.com",
            "AWS::NoValue": None,
            "AWS::Partition": "aws",
        }

    def __getitem__(self, key):
        resource_logical_id = key

        if resource_logical_id in self._parsed_resources:
            return self._parsed_resources[resource_logical_id]
        else:
            resource_json = self._resource_json_map.get(resource_logical_id)
            if not resource_json:
                raise KeyError(resource_logical_id)
            new_resource = parse_and_create_resource(
                resource_logical_id, resource_json, self, self._region_name)
            if new_resource is not None:
                self._parsed_resources[resource_logical_id] = new_resource
            return new_resource

    def __iter__(self):
        return iter(self.resources)

    def __len__(self):
        return len(self._resource_json_map)

    @property
    def resources(self):
        return self._resource_json_map.keys()

    def load_mapping(self):
        self._parsed_resources.update(self._template.get('Mappings', {}))

    def load_parameters(self):
        parameter_slots = self._template.get('Parameters', {})
        for parameter_name, parameter in parameter_slots.items():
            # Set the default values.
            self.resolved_parameters[parameter_name] = parameter.get('Default')

        # Set any input parameters that were passed
        self.no_echo_parameter_keys = []
        for key, value in self.input_parameters.items():
            if key in self.resolved_parameters:
                parameter_slot = parameter_slots[key]

                value_type = parameter_slot.get('Type', 'String')
                if value_type == 'CommaDelimitedList' or value_type.startswith("List"):
                    value = value.split(',')

                if parameter_slot.get('NoEcho'):
                    self.no_echo_parameter_keys.append(key)

                self.resolved_parameters[key] = value

        # Check if there are any non-default params that were not passed input
        # params
        for key, value in self.resolved_parameters.items():
            if value is None:
                raise MissingParameterError(key)

        self._parsed_resources.update(self.resolved_parameters)

    def load_conditions(self):
        conditions = self._template.get('Conditions', {})
        self.lazy_condition_map = LazyDict()
        for condition_name, condition in conditions.items():
            self.lazy_condition_map[condition_name] = functools.partial(parse_condition,
                condition, self._parsed_resources, self.lazy_condition_map)

        for condition_name in self.lazy_condition_map:
            self.lazy_condition_map[condition_name]

    def create(self):
        self.load_mapping()
        self.load_parameters()
        self.load_conditions()

        # Since this is a lazy map, to create every object we just need to
        # iterate through self.
        self.tags.update({'aws:cloudformation:stack-name': self.get('AWS::StackName'),
                          'aws:cloudformation:stack-id': self.get('AWS::StackId')})
        for resource in self.resources:
            if isinstance(self[resource], ec2_models.TaggedEC2Resource):
                self.tags['aws:cloudformation:logical-id'] = resource
                ec2_models.ec2_backends[self._region_name].create_tags(
                    [self[resource].physical_resource_id], self.tags)

    def diff(self, template, parameters=None):
        if parameters:
            self.input_parameters = parameters
        self.load_mapping()
        self.load_parameters()
        self.load_conditions()

        old_template = self._resource_json_map
        new_template = template['Resources']

        resource_names_by_action = {
            'Add': set(new_template) - set(old_template),
            'Modify': set(name for name in new_template if name in old_template and new_template[
                name] != old_template[name]),
            'Remove': set(old_template) - set(new_template)
        }
        resources_by_action = {
            'Add': {},
            'Modify': {},
            'Remove': {},
        }

        for resource_name in resource_names_by_action['Add']:
            resources_by_action['Add'][resource_name] = {
                'LogicalResourceId': resource_name,
                'ResourceType': new_template[resource_name]['Type']
            }

        for resource_name in resource_names_by_action['Modify']:
            resources_by_action['Modify'][resource_name] = {
                'LogicalResourceId': resource_name,
                'ResourceType': new_template[resource_name]['Type']
            }

        for resource_name in resource_names_by_action['Remove']:
            resources_by_action['Remove'][resource_name] = {
                'LogicalResourceId': resource_name,
                'ResourceType': old_template[resource_name]['Type']
            }

        return resources_by_action

    def update(self, template, parameters=None):
        resources_by_action = self.diff(template, parameters)

        old_template = self._resource_json_map
        new_template = template['Resources']
        self._resource_json_map = new_template

        for resource_name, resource in resources_by_action['Add'].items():
            resource_json = new_template[resource_name]
            new_resource = parse_and_create_resource(
                resource_name, resource_json, self, self._region_name)
            self._parsed_resources[resource_name] = new_resource

        for resource_name, resource in resources_by_action['Remove'].items():
            resource_json = old_template[resource_name]
            parse_and_delete_resource(
                resource_name, resource_json, self, self._region_name)
            self._parsed_resources.pop(resource_name)

        tries = 1
        while resources_by_action['Modify'] and tries < 5:
            for resource_name, resource in resources_by_action['Modify'].copy().items():
                resource_json = new_template[resource_name]
                try:
                    changed_resource = parse_and_update_resource(
                        resource_name, resource_json, self, self._region_name)
                except Exception as e:
                    # skip over dependency violations, and try again in a
                    # second pass
                    last_exception = e
                else:
                    self._parsed_resources[resource_name] = changed_resource
                    del resources_by_action['Modify'][resource_name]
            tries += 1
        if tries == 5:
            raise last_exception

    def delete(self):
        remaining_resources = set(self.resources)
        tries = 1
        while remaining_resources and tries < 5:
            for resource in remaining_resources.copy():
                parsed_resource = self._parsed_resources.get(resource)
                try:
                    if parsed_resource and hasattr(parsed_resource, 'delete'):
                        parsed_resource.delete(self._region_name)
                except Exception as e:
                    # skip over dependency violations, and try again in a
                    # second pass
                    last_exception = e
                else:
                    remaining_resources.remove(resource)
            tries += 1
        if tries == 5:
            raise last_exception


class OutputMap(collections.Mapping):

    def __init__(self, resources, template, stack_id):
        self._template = template
        self._stack_id = stack_id
        self._output_json_map = template.get('Outputs')

        # Create the default resources
        self._resource_map = resources
        self._parsed_outputs = dict()

    def __getitem__(self, key):
        output_logical_id = key

        if output_logical_id in self._parsed_outputs:
            return self._parsed_outputs[output_logical_id]
        else:
            output_json = self._output_json_map.get(output_logical_id)
            new_output = parse_output(
                output_logical_id, output_json, self._resource_map)
            self._parsed_outputs[output_logical_id] = new_output
            return new_output

    def __iter__(self):
        return iter(self.outputs)

    def __len__(self):
        return len(self._output_json_map)

    @property
    def outputs(self):
        return self._output_json_map.keys() if self._output_json_map else []

    @property
    def exports(self):
        exports = []
        if self.outputs:
            for key, value in self._output_json_map.items():
                if value.get('Export'):
                    cleaned_name = clean_json(value['Export'].get('Name'), self._resource_map)
                    cleaned_value = clean_json(value.get('Value'), self._resource_map)
                    exports.append(Export(self._stack_id, cleaned_name, cleaned_value))
        return exports

    def create(self):
        for output in self.outputs:
            self[output]


class Export(object):

    def __init__(self, exporting_stack_id, name, value):
        self._exporting_stack_id = exporting_stack_id
        self._name = name
        self._value = value

    @property
    def exporting_stack_id(self):
        return self._exporting_stack_id

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value



from __future__ import unicode_literals
import boto3
import botocore
import boto
import boto.ec2.elb
from boto.ec2.elb import HealthCheck
from boto.ec2.elb.attributes import (
    ConnectionSettingAttribute,
    ConnectionDrainingAttribute,
    AccessLogAttribute,
)
from botocore.exceptions import ClientError
from boto.exception import BotoServerError
from nose.tools import assert_raises
import sure  # noqa

from moto import mock_elb, mock_ec2, mock_elb_deprecated, mock_ec2_deprecated


@mock_elb_deprecated
@mock_ec2_deprecated
def test_create_load_balancer():
    conn = boto.connect_elb()
    ec2 = boto.ec2.connect_to_region("us-east-1")

    security_group = ec2.create_security_group('sg-abc987', 'description')

    zones = ['us-east-1a', 'us-east-1b']
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    conn.create_load_balancer('my-lb', zones, ports, scheme='internal', security_groups=[security_group.id])

    balancers = conn.get_all_load_balancers()
    balancer = balancers[0]
    balancer.name.should.equal("my-lb")
    balancer.scheme.should.equal("internal")
    list(balancer.security_groups).should.equal([security_group.id])
    set(balancer.availability_zones).should.equal(
        set(['us-east-1a', 'us-east-1b']))
    listener1 = balancer.listeners[0]
    listener1.load_balancer_port.should.equal(80)
    listener1.instance_port.should.equal(8080)
    listener1.protocol.should.equal("HTTP")
    listener2 = balancer.listeners[1]
    listener2.load_balancer_port.should.equal(443)
    listener2.instance_port.should.equal(8443)
    listener2.protocol.should.equal("TCP")


@mock_elb_deprecated
def test_getting_missing_elb():
    conn = boto.connect_elb()
    conn.get_all_load_balancers.when.called_with(
        load_balancer_names='aaa').should.throw(BotoServerError)


@mock_elb_deprecated
def test_create_elb_in_multiple_region():
    zones = ['us-east-1a', 'us-east-1b']
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]

    west1_conn = boto.ec2.elb.connect_to_region("us-west-1")
    west1_conn.create_load_balancer('my-lb', zones, ports)

    west2_conn = boto.ec2.elb.connect_to_region("us-west-2")
    west2_conn.create_load_balancer('my-lb', zones, ports)

    list(west1_conn.get_all_load_balancers()).should.have.length_of(1)
    list(west2_conn.get_all_load_balancers()).should.have.length_of(1)


@mock_elb_deprecated
def test_create_load_balancer_with_certificate():
    conn = boto.connect_elb()

    zones = ['us-east-1a']
    ports = [
        (443, 8443, 'https', 'arn:aws:iam:123456789012:server-certificate/test-cert')]
    conn.create_load_balancer('my-lb', zones, ports)

    balancers = conn.get_all_load_balancers()
    balancer = balancers[0]
    balancer.name.should.equal("my-lb")
    balancer.scheme.should.equal("internet-facing")
    set(balancer.availability_zones).should.equal(set(['us-east-1a']))
    listener = balancer.listeners[0]
    listener.load_balancer_port.should.equal(443)
    listener.instance_port.should.equal(8443)
    listener.protocol.should.equal("HTTPS")
    listener.ssl_certificate_id.should.equal(
        'arn:aws:iam:123456789012:server-certificate/test-cert')


@mock_elb
def test_create_and_delete_boto3_support():
    client = boto3.client('elb', region_name='us-east-1')

    client.create_load_balancer(
        LoadBalancerName='my-lb',
        Listeners=[
            {'Protocol': 'tcp', 'LoadBalancerPort': 80, 'InstancePort': 8080}],
        AvailabilityZones=['us-east-1a', 'us-east-1b']
    )
    list(client.describe_load_balancers()[
         'LoadBalancerDescriptions']).should.have.length_of(1)

    client.delete_load_balancer(
        LoadBalancerName='my-lb'
    )
    list(client.describe_load_balancers()[
         'LoadBalancerDescriptions']).should.have.length_of(0)


@mock_elb
def test_create_load_balancer_with_no_listeners_defined():
    client = boto3.client('elb', region_name='us-east-1')

    with assert_raises(ClientError):
        client.create_load_balancer(
            LoadBalancerName='my-lb',
            Listeners=[],
            AvailabilityZones=['us-east-1a', 'us-east-1b']
        )


@mock_elb
def test_describe_paginated_balancers():
    client = boto3.client('elb', region_name='us-east-1')

    for i in range(51):
        client.create_load_balancer(
            LoadBalancerName='my-lb%d' % i,
            Listeners=[
                {'Protocol': 'tcp', 'LoadBalancerPort': 80, 'InstancePort': 8080}],
            AvailabilityZones=['us-east-1a', 'us-east-1b']
        )

    resp = client.describe_load_balancers()
    resp['LoadBalancerDescriptions'].should.have.length_of(50)
    resp['NextMarker'].should.equal(resp['LoadBalancerDescriptions'][-1]['LoadBalancerName'])
    resp2 = client.describe_load_balancers(Marker=resp['NextMarker'])
    resp2['LoadBalancerDescriptions'].should.have.length_of(1)
    assert 'NextToken' not in resp2.keys()


@mock_elb
@mock_ec2
def test_apply_security_groups_to_load_balancer():
    client = boto3.client('elb', region_name='us-east-1')
    ec2 = boto3.resource('ec2', region_name='us-east-1')

    vpc = ec2.create_vpc(CidrBlock='10.0.0.0/16')
    security_group = ec2.create_security_group(
        GroupName='sg01', Description='Test security group sg01', VpcId=vpc.id)

    client.create_load_balancer(
        LoadBalancerName='my-lb',
        Listeners=[
            {'Protocol': 'tcp', 'LoadBalancerPort': 80, 'InstancePort': 8080}],
        AvailabilityZones=['us-east-1a', 'us-east-1b']
    )

    response = client.apply_security_groups_to_load_balancer(
        LoadBalancerName='my-lb',
        SecurityGroups=[security_group.id])

    assert response['SecurityGroups'] == [security_group.id]
    balancer = client.describe_load_balancers()['LoadBalancerDescriptions'][0]
    assert balancer['SecurityGroups'] == [security_group.id]

    # Using a not-real security group raises an error
    with assert_raises(ClientError) as error:
        response = client.apply_security_groups_to_load_balancer(
            LoadBalancerName='my-lb',
            SecurityGroups=['not-really-a-security-group'])
    assert "One or more of the specified security groups do not exist." in str(error.exception)


@mock_elb_deprecated
def test_add_listener():
    conn = boto.connect_elb()
    zones = ['us-east-1a', 'us-east-1b']
    ports = [(80, 8080, 'http')]
    conn.create_load_balancer('my-lb', zones, ports)
    new_listener = (443, 8443, 'tcp')
    conn.create_load_balancer_listeners('my-lb', [new_listener])
    balancers = conn.get_all_load_balancers()
    balancer = balancers[0]
    listener1 = balancer.listeners[0]
    listener1.load_balancer_port.should.equal(80)
    listener1.instance_port.should.equal(8080)
    listener1.protocol.should.equal("HTTP")
    listener2 = balancer.listeners[1]
    listener2.load_balancer_port.should.equal(443)
    listener2.instance_port.should.equal(8443)
    listener2.protocol.should.equal("TCP")


@mock_elb_deprecated
def test_delete_listener():
    conn = boto.connect_elb()

    zones = ['us-east-1a', 'us-east-1b']
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    conn.create_load_balancer('my-lb', zones, ports)
    conn.delete_load_balancer_listeners('my-lb', [443])
    balancers = conn.get_all_load_balancers()
    balancer = balancers[0]
    listener1 = balancer.listeners[0]
    listener1.load_balancer_port.should.equal(80)
    listener1.instance_port.should.equal(8080)
    listener1.protocol.should.equal("HTTP")
    balancer.listeners.should.have.length_of(1)


@mock_elb
def test_create_and_delete_listener_boto3_support():
    client = boto3.client('elb', region_name='us-east-1')

    client.create_load_balancer(
        LoadBalancerName='my-lb',
        Listeners=[{'Protocol': 'http',
                    'LoadBalancerPort': 80, 'InstancePort': 8080}],
        AvailabilityZones=['us-east-1a', 'us-east-1b']
    )
    list(client.describe_load_balancers()[
         'LoadBalancerDescriptions']).should.have.length_of(1)

    client.create_load_balancer_listeners(
        LoadBalancerName='my-lb',
        Listeners=[
            {'Protocol': 'tcp', 'LoadBalancerPort': 443, 'InstancePort': 8443}]
    )
    balancer = client.describe_load_balancers()['LoadBalancerDescriptions'][0]
    list(balancer['ListenerDescriptions']).should.have.length_of(2)
    balancer['ListenerDescriptions'][0][
        'Listener']['Protocol'].should.equal('HTTP')
    balancer['ListenerDescriptions'][0]['Listener'][
        'LoadBalancerPort'].should.equal(80)
    balancer['ListenerDescriptions'][0]['Listener'][
        'InstancePort'].should.equal(8080)
    balancer['ListenerDescriptions'][1][
        'Listener']['Protocol'].should.equal('TCP')
    balancer['ListenerDescriptions'][1]['Listener'][
        'LoadBalancerPort'].should.equal(443)
    balancer['ListenerDescriptions'][1]['Listener'][
        'InstancePort'].should.equal(8443)

    # Creating this listener with an conflicting definition throws error
    with assert_raises(ClientError):
        client.create_load_balancer_listeners(
            LoadBalancerName='my-lb',
            Listeners=[
                {'Protocol': 'tcp', 'LoadBalancerPort': 443, 'InstancePort': 1234}]
        )

    client.delete_load_balancer_listeners(
        LoadBalancerName='my-lb',
        LoadBalancerPorts=[443])

    balancer = client.describe_load_balancers()['LoadBalancerDescriptions'][0]
    list(balancer['ListenerDescriptions']).should.have.length_of(1)


@mock_elb_deprecated
def test_set_sslcertificate():
    conn = boto.connect_elb()

    zones = ['us-east-1a', 'us-east-1b']
    ports = [(443, 8443, 'tcp')]
    conn.create_load_balancer('my-lb', zones, ports)
    conn.set_lb_listener_SSL_certificate('my-lb', '443', 'arn:certificate')
    balancers = conn.get_all_load_balancers()
    balancer = balancers[0]
    listener1 = balancer.listeners[0]
    listener1.load_balancer_port.should.equal(443)
    listener1.instance_port.should.equal(8443)
    listener1.protocol.should.equal("TCP")
    listener1.ssl_certificate_id.should.equal("arn:certificate")


@mock_elb_deprecated
def test_get_load_balancers_by_name():
    conn = boto.connect_elb()

    zones = ['us-east-1a', 'us-east-1b']
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    conn.create_load_balancer('my-lb1', zones, ports)
    conn.create_load_balancer('my-lb2', zones, ports)
    conn.create_load_balancer('my-lb3', zones, ports)

    conn.get_all_load_balancers().should.have.length_of(3)
    conn.get_all_load_balancers(
        load_balancer_names=['my-lb1']).should.have.length_of(1)
    conn.get_all_load_balancers(
        load_balancer_names=['my-lb1', 'my-lb2']).should.have.length_of(2)


@mock_elb_deprecated
def test_delete_load_balancer():
    conn = boto.connect_elb()

    zones = ['us-east-1a']
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    conn.create_load_balancer('my-lb', zones, ports)

    balancers = conn.get_all_load_balancers()
    balancers.should.have.length_of(1)

    conn.delete_load_balancer("my-lb")
    balancers = conn.get_all_load_balancers()
    balancers.should.have.length_of(0)


@mock_elb_deprecated
def test_create_health_check():
    conn = boto.connect_elb()

    hc = HealthCheck(
        interval=20,
        healthy_threshold=3,
        unhealthy_threshold=5,
        target='HTTP:8080/health',
        timeout=23,
    )

    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)
    lb.configure_health_check(hc)

    balancer = conn.get_all_load_balancers()[0]
    health_check = balancer.health_check
    health_check.interval.should.equal(20)
    health_check.healthy_threshold.should.equal(3)
    health_check.unhealthy_threshold.should.equal(5)
    health_check.target.should.equal('HTTP:8080/health')
    health_check.timeout.should.equal(23)


@mock_elb
def test_create_health_check_boto3():
    client = boto3.client('elb', region_name='us-east-1')

    client.create_load_balancer(
        LoadBalancerName='my-lb',
        Listeners=[{'Protocol': 'http',
                    'LoadBalancerPort': 80, 'InstancePort': 8080}],
        AvailabilityZones=['us-east-1a', 'us-east-1b']
    )
    client.configure_health_check(
        LoadBalancerName='my-lb',
        HealthCheck={
            'Target': 'HTTP:8080/health',
            'Interval': 20,
            'Timeout': 23,
            'HealthyThreshold': 3,
            'UnhealthyThreshold': 5
        }
    )

    balancer = client.describe_load_balancers()['LoadBalancerDescriptions'][0]
    balancer['HealthCheck']['Target'].should.equal('HTTP:8080/health')
    balancer['HealthCheck']['Interval'].should.equal(20)
    balancer['HealthCheck']['Timeout'].should.equal(23)
    balancer['HealthCheck']['HealthyThreshold'].should.equal(3)
    balancer['HealthCheck']['UnhealthyThreshold'].should.equal(5)


@mock_ec2_deprecated
@mock_elb_deprecated
def test_register_instances():
    ec2_conn = boto.connect_ec2()
    reservation = ec2_conn.run_instances('ami-1234abcd', 2)
    instance_id1 = reservation.instances[0].id
    instance_id2 = reservation.instances[1].id

    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)

    lb.register_instances([instance_id1, instance_id2])

    balancer = conn.get_all_load_balancers()[0]
    instance_ids = [instance.id for instance in balancer.instances]
    set(instance_ids).should.equal(set([instance_id1, instance_id2]))


@mock_ec2
@mock_elb
def test_register_instances_boto3():
    ec2 = boto3.resource('ec2', region_name='us-east-1')
    response = ec2.create_instances(
        ImageId='ami-1234abcd', MinCount=2, MaxCount=2)
    instance_id1 = response[0].id
    instance_id2 = response[1].id

    client = boto3.client('elb', region_name='us-east-1')
    client.create_load_balancer(
        LoadBalancerName='my-lb',
        Listeners=[{'Protocol': 'http',
                    'LoadBalancerPort': 80, 'InstancePort': 8080}],
        AvailabilityZones=['us-east-1a', 'us-east-1b']
    )
    client.register_instances_with_load_balancer(
        LoadBalancerName='my-lb',
        Instances=[
            {'InstanceId': instance_id1},
            {'InstanceId': instance_id2}
        ]
    )
    balancer = client.describe_load_balancers()['LoadBalancerDescriptions'][0]
    instance_ids = [instance['InstanceId']
                    for instance in balancer['Instances']]
    set(instance_ids).should.equal(set([instance_id1, instance_id2]))


@mock_ec2_deprecated
@mock_elb_deprecated
def test_deregister_instances():
    ec2_conn = boto.connect_ec2()
    reservation = ec2_conn.run_instances('ami-1234abcd', 2)
    instance_id1 = reservation.instances[0].id
    instance_id2 = reservation.instances[1].id

    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)

    lb.register_instances([instance_id1, instance_id2])

    balancer = conn.get_all_load_balancers()[0]
    balancer.instances.should.have.length_of(2)
    balancer.deregister_instances([instance_id1])

    balancer.instances.should.have.length_of(1)
    balancer.instances[0].id.should.equal(instance_id2)


@mock_ec2
@mock_elb
def test_deregister_instances_boto3():
    ec2 = boto3.resource('ec2', region_name='us-east-1')
    response = ec2.create_instances(
        ImageId='ami-1234abcd', MinCount=2, MaxCount=2)
    instance_id1 = response[0].id
    instance_id2 = response[1].id

    client = boto3.client('elb', region_name='us-east-1')
    client.create_load_balancer(
        LoadBalancerName='my-lb',
        Listeners=[{'Protocol': 'http',
                    'LoadBalancerPort': 80, 'InstancePort': 8080}],
        AvailabilityZones=['us-east-1a', 'us-east-1b']
    )
    client.register_instances_with_load_balancer(
        LoadBalancerName='my-lb',
        Instances=[
            {'InstanceId': instance_id1},
            {'InstanceId': instance_id2}
        ]
    )

    balancer = client.describe_load_balancers()['LoadBalancerDescriptions'][0]
    balancer['Instances'].should.have.length_of(2)

    client.deregister_instances_from_load_balancer(
        LoadBalancerName='my-lb',
        Instances=[
            {'InstanceId': instance_id1}
        ]
    )

    balancer = client.describe_load_balancers()['LoadBalancerDescriptions'][0]
    balancer['Instances'].should.have.length_of(1)
    balancer['Instances'][0]['InstanceId'].should.equal(instance_id2)


@mock_elb_deprecated
def test_default_attributes():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)
    attributes = lb.get_attributes()

    attributes.cross_zone_load_balancing.enabled.should.be.false
    attributes.connection_draining.enabled.should.be.false
    attributes.access_log.enabled.should.be.false
    attributes.connecting_settings.idle_timeout.should.equal(60)


@mock_elb_deprecated
def test_cross_zone_load_balancing_attribute():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)

    conn.modify_lb_attribute("my-lb", "CrossZoneLoadBalancing", True)
    attributes = lb.get_attributes(force=True)
    attributes.cross_zone_load_balancing.enabled.should.be.true

    conn.modify_lb_attribute("my-lb", "CrossZoneLoadBalancing", False)
    attributes = lb.get_attributes(force=True)
    attributes.cross_zone_load_balancing.enabled.should.be.false


@mock_elb_deprecated
def test_connection_draining_attribute():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)

    connection_draining = ConnectionDrainingAttribute()
    connection_draining.enabled = True
    connection_draining.timeout = 60

    conn.modify_lb_attribute(
        "my-lb", "ConnectionDraining", connection_draining)
    attributes = lb.get_attributes(force=True)
    attributes.connection_draining.enabled.should.be.true
    attributes.connection_draining.timeout.should.equal(60)

    connection_draining.timeout = 30
    conn.modify_lb_attribute(
        "my-lb", "ConnectionDraining", connection_draining)
    attributes = lb.get_attributes(force=True)
    attributes.connection_draining.timeout.should.equal(30)

    connection_draining.enabled = False
    conn.modify_lb_attribute(
        "my-lb", "ConnectionDraining", connection_draining)
    attributes = lb.get_attributes(force=True)
    attributes.connection_draining.enabled.should.be.false


@mock_elb_deprecated
def test_access_log_attribute():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)

    access_log = AccessLogAttribute()
    access_log.enabled = True
    access_log.s3_bucket_name = 'bucket'
    access_log.s3_bucket_prefix = 'prefix'
    access_log.emit_interval = 60

    conn.modify_lb_attribute("my-lb", "AccessLog", access_log)
    attributes = lb.get_attributes(force=True)
    attributes.access_log.enabled.should.be.true
    attributes.access_log.s3_bucket_name.should.equal("bucket")
    attributes.access_log.s3_bucket_prefix.should.equal("prefix")
    attributes.access_log.emit_interval.should.equal(60)

    access_log.enabled = False
    conn.modify_lb_attribute("my-lb", "AccessLog", access_log)
    attributes = lb.get_attributes(force=True)
    attributes.access_log.enabled.should.be.false


@mock_elb_deprecated
def test_connection_settings_attribute():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)

    connection_settings = ConnectionSettingAttribute(conn)
    connection_settings.idle_timeout = 120

    conn.modify_lb_attribute(
        "my-lb", "ConnectingSettings", connection_settings)
    attributes = lb.get_attributes(force=True)
    attributes.connecting_settings.idle_timeout.should.equal(120)

    connection_settings.idle_timeout = 60
    conn.modify_lb_attribute(
        "my-lb", "ConnectingSettings", connection_settings)
    attributes = lb.get_attributes(force=True)
    attributes.connecting_settings.idle_timeout.should.equal(60)


@mock_elb_deprecated
def test_create_lb_cookie_stickiness_policy():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)
    cookie_expiration_period = 60
    policy_name = "LBCookieStickinessPolicy"

    lb.create_cookie_stickiness_policy(cookie_expiration_period, policy_name)

    lb = conn.get_all_load_balancers()[0]
    # There appears to be a quirk about boto, whereby it returns a unicode
    # string for cookie_expiration_period, despite being stated in
    # documentation to be a long numeric.
    #
    # To work around that, this value is converted to an int and checked.
    cookie_expiration_period_response_str = lb.policies.lb_cookie_stickiness_policies[
        0].cookie_expiration_period
    int(cookie_expiration_period_response_str).should.equal(
        cookie_expiration_period)
    lb.policies.lb_cookie_stickiness_policies[
        0].policy_name.should.equal(policy_name)


@mock_elb_deprecated
def test_create_lb_cookie_stickiness_policy_no_expiry():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)
    policy_name = "LBCookieStickinessPolicy"

    lb.create_cookie_stickiness_policy(None, policy_name)

    lb = conn.get_all_load_balancers()[0]
    lb.policies.lb_cookie_stickiness_policies[
        0].cookie_expiration_period.should.be.none
    lb.policies.lb_cookie_stickiness_policies[
        0].policy_name.should.equal(policy_name)


@mock_elb_deprecated
def test_create_app_cookie_stickiness_policy():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)
    cookie_name = "my-stickiness-policy"
    policy_name = "AppCookieStickinessPolicy"

    lb.create_app_cookie_stickiness_policy(cookie_name, policy_name)

    lb = conn.get_all_load_balancers()[0]
    lb.policies.app_cookie_stickiness_policies[
        0].cookie_name.should.equal(cookie_name)
    lb.policies.app_cookie_stickiness_policies[
        0].policy_name.should.equal(policy_name)


@mock_elb_deprecated
def test_create_lb_policy():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)
    policy_name = "ProxyPolicy"

    lb.create_lb_policy(policy_name, 'ProxyProtocolPolicyType', {
                        'ProxyProtocol': True})

    lb = conn.get_all_load_balancers()[0]
    lb.policies.other_policies[0].policy_name.should.equal(policy_name)


@mock_elb_deprecated
def test_set_policies_of_listener():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)
    listener_port = 80
    policy_name = "my-stickiness-policy"

    # boto docs currently state that zero or one policy may be associated
    # with a given listener

    # in a real flow, it is necessary first to create a policy,
    # then to set that policy to the listener
    lb.create_cookie_stickiness_policy(None, policy_name)
    lb.set_policies_of_listener(listener_port, [policy_name])

    lb = conn.get_all_load_balancers()[0]
    listener = lb.listeners[0]
    listener.load_balancer_port.should.equal(listener_port)
    # by contrast to a backend, a listener stores only policy name strings
    listener.policy_names[0].should.equal(policy_name)


@mock_elb_deprecated
def test_set_policies_of_backend_server():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', [], ports)
    instance_port = 8080
    policy_name = "ProxyPolicy"

    # in a real flow, it is necessary first to create a policy,
    # then to set that policy to the backend
    lb.create_lb_policy(policy_name, 'ProxyProtocolPolicyType', {
                        'ProxyProtocol': True})
    lb.set_policies_of_backend_server(instance_port, [policy_name])

    lb = conn.get_all_load_balancers()[0]
    backend = lb.backends[0]
    backend.instance_port.should.equal(instance_port)
    # by contrast to a listener, a backend stores OtherPolicy objects
    backend.policies[0].policy_name.should.equal(policy_name)


@mock_ec2_deprecated
@mock_elb_deprecated
def test_describe_instance_health():
    ec2_conn = boto.connect_ec2()
    reservation = ec2_conn.run_instances('ami-1234abcd', 2)
    instance_id1 = reservation.instances[0].id
    instance_id2 = reservation.instances[1].id

    conn = boto.connect_elb()
    zones = ['us-east-1a', 'us-east-1b']
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    lb = conn.create_load_balancer('my-lb', zones, ports)

    instances_health = conn.describe_instance_health('my-lb')
    instances_health.should.be.empty

    lb.register_instances([instance_id1, instance_id2])

    instances_health = conn.describe_instance_health('my-lb')
    instances_health.should.have.length_of(2)
    for instance_health in instances_health:
        instance_health.instance_id.should.be.within(
            [instance_id1, instance_id2])
        instance_health.state.should.equal('InService')

    instances_health = conn.describe_instance_health('my-lb', [instance_id1])
    instances_health.should.have.length_of(1)
    instances_health[0].instance_id.should.equal(instance_id1)
    instances_health[0].state.should.equal('InService')


@mock_ec2
@mock_elb
def test_describe_instance_health_boto3():
    elb = boto3.client('elb', region_name="us-east-1")
    ec2 = boto3.client('ec2', region_name="us-east-1")
    instances = ec2.run_instances(MinCount=2, MaxCount=2)['Instances']
    lb_name = "my_load_balancer"
    elb.create_load_balancer(
        Listeners=[{
            'InstancePort': 80,
            'LoadBalancerPort': 8080,
            'Protocol': 'HTTP'
        }],
        LoadBalancerName=lb_name,
    )
    elb.register_instances_with_load_balancer(
        LoadBalancerName=lb_name,
        Instances=[{'InstanceId': instances[0]['InstanceId']}]
    )
    instances_health = elb.describe_instance_health(
        LoadBalancerName=lb_name,
        Instances=[{'InstanceId': instance['InstanceId']} for instance in instances]
    )
    instances_health['InstanceStates'].should.have.length_of(2)
    instances_health['InstanceStates'][0]['InstanceId'].\
        should.equal(instances[0]['InstanceId'])
    instances_health['InstanceStates'][0]['State'].\
        should.equal('InService')
    instances_health['InstanceStates'][1]['InstanceId'].\
        should.equal(instances[1]['InstanceId'])
    instances_health['InstanceStates'][1]['State'].\
        should.equal('Unknown')


@mock_elb
def test_add_remove_tags():
    client = boto3.client('elb', region_name='us-east-1')

    client.add_tags.when.called_with(LoadBalancerNames=['my-lb'],
                                     Tags=[{
                                         'Key': 'a',
                                         'Value': 'b'
                                     }]).should.throw(botocore.exceptions.ClientError)

    client.create_load_balancer(
        LoadBalancerName='my-lb',
        Listeners=[
            {'Protocol': 'tcp', 'LoadBalancerPort': 80, 'InstancePort': 8080}],
        AvailabilityZones=['us-east-1a', 'us-east-1b']
    )

    list(client.describe_load_balancers()[
         'LoadBalancerDescriptions']).should.have.length_of(1)

    client.add_tags(LoadBalancerNames=['my-lb'],
                    Tags=[{
                        'Key': 'a',
                        'Value': 'b'
                    }])

    tags = dict([(d['Key'], d['Value']) for d in client.describe_tags(
        LoadBalancerNames=['my-lb'])['TagDescriptions'][0]['Tags']])
    tags.should.have.key('a').which.should.equal('b')

    client.add_tags(LoadBalancerNames=['my-lb'],
                    Tags=[{
                        'Key': 'a',
                        'Value': 'b'
                    }, {
                        'Key': 'b',
                        'Value': 'b'
                    }, {
                        'Key': 'c',
                        'Value': 'b'
                    }, {
                        'Key': 'd',
                        'Value': 'b'
                    }, {
                        'Key': 'e',
                        'Value': 'b'
                    }, {
                        'Key': 'f',
                        'Value': 'b'
                    }, {
                        'Key': 'g',
                        'Value': 'b'
                    }, {
                        'Key': 'h',
                        'Value': 'b'
                    }, {
                        'Key': 'i',
                        'Value': 'b'
                    }, {
                        'Key': 'j',
                        'Value': 'b'
                    }])

    client.add_tags.when.called_with(LoadBalancerNames=['my-lb'],
                                     Tags=[{
                                         'Key': 'k',
                                         'Value': 'b'
                                     }]).should.throw(botocore.exceptions.ClientError)

    client.add_tags(LoadBalancerNames=['my-lb'],
                    Tags=[{
                        'Key': 'j',
                        'Value': 'c'
                    }])

    tags = dict([(d['Key'], d['Value']) for d in client.describe_tags(
        LoadBalancerNames=['my-lb'])['TagDescriptions'][0]['Tags']])

    tags.should.have.key('a').which.should.equal('b')
    tags.should.have.key('b').which.should.equal('b')
    tags.should.have.key('c').which.should.equal('b')
    tags.should.have.key('d').which.should.equal('b')
    tags.should.have.key('e').which.should.equal('b')
    tags.should.have.key('f').which.should.equal('b')
    tags.should.have.key('g').which.should.equal('b')
    tags.should.have.key('h').which.should.equal('b')
    tags.should.have.key('i').which.should.equal('b')
    tags.should.have.key('j').which.should.equal('c')
    tags.shouldnt.have.key('k')

    client.remove_tags(LoadBalancerNames=['my-lb'],
                       Tags=[{
                           'Key': 'a'
                       }])

    tags = dict([(d['Key'], d['Value']) for d in client.describe_tags(
        LoadBalancerNames=['my-lb'])['TagDescriptions'][0]['Tags']])

    tags.shouldnt.have.key('a')
    tags.should.have.key('b').which.should.equal('b')
    tags.should.have.key('c').which.should.equal('b')
    tags.should.have.key('d').which.should.equal('b')
    tags.should.have.key('e').which.should.equal('b')
    tags.should.have.key('f').which.should.equal('b')
    tags.should.have.key('g').which.should.equal('b')
    tags.should.have.key('h').which.should.equal('b')
    tags.should.have.key('i').which.should.equal('b')
    tags.should.have.key('j').which.should.equal('c')

    client.create_load_balancer(
        LoadBalancerName='other-lb',
        Listeners=[
            {'Protocol': 'tcp', 'LoadBalancerPort': 433, 'InstancePort': 8433}],
        AvailabilityZones=['us-east-1a', 'us-east-1b']
    )

    client.add_tags(LoadBalancerNames=['other-lb'],
                    Tags=[{
                        'Key': 'other',
                        'Value': 'something'
                    }])

    lb_tags = dict([(l['LoadBalancerName'], dict([(d['Key'], d['Value']) for d in l['Tags']]))
                    for l in client.describe_tags(LoadBalancerNames=['my-lb', 'other-lb'])['TagDescriptions']])

    lb_tags.should.have.key('my-lb')
    lb_tags.should.have.key('other-lb')

    lb_tags['my-lb'].shouldnt.have.key('other')
    lb_tags[
        'other-lb'].should.have.key('other').which.should.equal('something')


@mock_elb
def test_create_with_tags():
    client = boto3.client('elb', region_name='us-east-1')

    client.create_load_balancer(
        LoadBalancerName='my-lb',
        Listeners=[
            {'Protocol': 'tcp', 'LoadBalancerPort': 80, 'InstancePort': 8080}],
        AvailabilityZones=['us-east-1a', 'us-east-1b'],
        Tags=[{
            'Key': 'k',
            'Value': 'v'
        }]
    )

    tags = dict((d['Key'], d['Value']) for d in client.describe_tags(
        LoadBalancerNames=['my-lb'])['TagDescriptions'][0]['Tags'])
    tags.should.have.key('k').which.should.equal('v')


@mock_elb
def test_modify_attributes():
    client = boto3.client('elb', region_name='us-east-1')

    client.create_load_balancer(
        LoadBalancerName='my-lb',
        Listeners=[{'Protocol': 'tcp', 'LoadBalancerPort': 80, 'InstancePort': 8080}],
        AvailabilityZones=['us-east-1a', 'us-east-1b']
    )

    # Default ConnectionDraining timeout of 300 seconds
    client.modify_load_balancer_attributes(
        LoadBalancerName='my-lb',
        LoadBalancerAttributes={
            'ConnectionDraining': {'Enabled': True},
        }
    )
    lb_attrs = client.describe_load_balancer_attributes(LoadBalancerName='my-lb')
    lb_attrs['LoadBalancerAttributes']['ConnectionDraining']['Enabled'].should.equal(True)
    lb_attrs['LoadBalancerAttributes']['ConnectionDraining']['Timeout'].should.equal(300)

    # specify a custom ConnectionDraining timeout
    client.modify_load_balancer_attributes(
        LoadBalancerName='my-lb',
        LoadBalancerAttributes={
            'ConnectionDraining': {
                'Enabled': True,
                'Timeout': 45,
            },
        }
    )
    lb_attrs = client.describe_load_balancer_attributes(LoadBalancerName='my-lb')
    lb_attrs['LoadBalancerAttributes']['ConnectionDraining']['Enabled'].should.equal(True)
    lb_attrs['LoadBalancerAttributes']['ConnectionDraining']['Timeout'].should.equal(45)


@mock_ec2
@mock_elb
def test_subnets():
    ec2 = boto3.resource('ec2', region_name='us-east-1')
    vpc = ec2.create_vpc(
        CidrBlock='172.28.7.0/24',
        InstanceTenancy='default'
    )
    subnet = ec2.create_subnet(
        VpcId=vpc.id,
        CidrBlock='172.28.7.192/26'
    )
    client = boto3.client('elb', region_name='us-east-1')
    client.create_load_balancer(
        LoadBalancerName='my-lb',
        Listeners=[
            {'Protocol': 'tcp', 'LoadBalancerPort': 80, 'InstancePort': 8080}],
        Subnets=[subnet.id]
    )

    lb = client.describe_load_balancers()['LoadBalancerDescriptions'][0]
    lb.should.have.key('Subnets').which.should.have.length_of(1)
    lb['Subnets'][0].should.equal(subnet.id)

    lb.should.have.key('VPCId').which.should.equal(vpc.id)


@mock_elb_deprecated
def test_create_load_balancer_duplicate():
    conn = boto.connect_elb()
    ports = [(80, 8080, 'http'), (443, 8443, 'tcp')]
    conn.create_load_balancer('my-lb', [], ports)
    conn.create_load_balancer.when.called_with(
        'my-lb', [], ports).should.throw(BotoServerError)



from __future__ import unicode_literals

import json
import sure  # noqa
import boto3

from moto import mock_iot
from botocore.exceptions import ClientError
from nose.tools import assert_raises


@mock_iot
def test_things():
    client = boto3.client('iot', region_name='ap-northeast-1')
    name = 'my-thing'
    type_name = 'my-type-name'

    # thing type
    thing_type = client.create_thing_type(thingTypeName=type_name)
    thing_type.should.have.key('thingTypeName').which.should.equal(type_name)
    thing_type.should.have.key('thingTypeArn')

    res = client.list_thing_types()
    res.should.have.key('thingTypes').which.should.have.length_of(1)
    for thing_type in res['thingTypes']:
        thing_type.should.have.key('thingTypeName').which.should_not.be.none

    thing_type = client.describe_thing_type(thingTypeName=type_name)
    thing_type.should.have.key('thingTypeName').which.should.equal(type_name)
    thing_type.should.have.key('thingTypeProperties')
    thing_type.should.have.key('thingTypeMetadata')

    # thing
    thing = client.create_thing(thingName=name, thingTypeName=type_name)
    thing.should.have.key('thingName').which.should.equal(name)
    thing.should.have.key('thingArn')
    res = client.list_things()
    res.should.have.key('things').which.should.have.length_of(1)
    for thing in res['things']:
        thing.should.have.key('thingName').which.should_not.be.none
        thing.should.have.key('thingArn').which.should_not.be.none

    thing = client.update_thing(thingName=name, attributePayload={'attributes': {'k1': 'v1'}})
    res = client.list_things()
    res.should.have.key('things').which.should.have.length_of(1)
    for thing in res['things']:
        thing.should.have.key('thingName').which.should_not.be.none
        thing.should.have.key('thingArn').which.should_not.be.none
    res['things'][0]['attributes'].should.have.key('k1').which.should.equal('v1')

    thing = client.describe_thing(thingName=name)
    thing.should.have.key('thingName').which.should.equal(name)
    thing.should.have.key('defaultClientId')
    thing.should.have.key('thingTypeName')
    thing.should.have.key('attributes')
    thing.should.have.key('version')

    # delete thing
    client.delete_thing(thingName=name)
    res = client.list_things()
    res.should.have.key('things').which.should.have.length_of(0)

    # delete thing type
    client.delete_thing_type(thingTypeName=type_name)
    res = client.list_thing_types()
    res.should.have.key('thingTypes').which.should.have.length_of(0)


@mock_iot
def test_list_thing_types():
    client = boto3.client('iot', region_name='ap-northeast-1')

    for i in range(0, 100):
        client.create_thing_type(thingTypeName=str(i + 1))

    thing_types = client.list_thing_types()
    thing_types.should.have.key('nextToken')
    thing_types.should.have.key('thingTypes').which.should.have.length_of(50)
    thing_types['thingTypes'][0]['thingTypeName'].should.equal('1')
    thing_types['thingTypes'][-1]['thingTypeName'].should.equal('50')

    thing_types = client.list_thing_types(nextToken=thing_types['nextToken'])
    thing_types.should.have.key('thingTypes').which.should.have.length_of(50)
    thing_types.should_not.have.key('nextToken')
    thing_types['thingTypes'][0]['thingTypeName'].should.equal('51')
    thing_types['thingTypes'][-1]['thingTypeName'].should.equal('100')


@mock_iot
def test_list_thing_types_with_typename_filter():
    client = boto3.client('iot', region_name='ap-northeast-1')

    client.create_thing_type(thingTypeName='thing')
    client.create_thing_type(thingTypeName='thingType')
    client.create_thing_type(thingTypeName='thingTypeName')
    client.create_thing_type(thingTypeName='thingTypeNameGroup')
    client.create_thing_type(thingTypeName='shouldNotFind')
    client.create_thing_type(thingTypeName='find me it shall not')

    thing_types = client.list_thing_types(thingTypeName='thing')
    thing_types.should_not.have.key('nextToken')
    thing_types.should.have.key('thingTypes').which.should.have.length_of(4)
    thing_types['thingTypes'][0]['thingTypeName'].should.equal('thing')
    thing_types['thingTypes'][-1]['thingTypeName'].should.equal('thingTypeNameGroup')

    thing_types = client.list_thing_types(thingTypeName='thingTypeName')
    thing_types.should_not.have.key('nextToken')
    thing_types.should.have.key('thingTypes').which.should.have.length_of(2)
    thing_types['thingTypes'][0]['thingTypeName'].should.equal('thingTypeName')
    thing_types['thingTypes'][-1]['thingTypeName'].should.equal('thingTypeNameGroup')


@mock_iot
def test_list_things_with_next_token():
    client = boto3.client('iot', region_name='ap-northeast-1')

    for i in range(0, 200):
        client.create_thing(thingName=str(i + 1))

    things = client.list_things()
    things.should.have.key('nextToken')
    things.should.have.key('things').which.should.have.length_of(50)
    things['things'][0]['thingName'].should.equal('1')
    things['things'][0]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/1')
    things['things'][-1]['thingName'].should.equal('50')
    things['things'][-1]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/50')

    things = client.list_things(nextToken=things['nextToken'])
    things.should.have.key('nextToken')
    things.should.have.key('things').which.should.have.length_of(50)
    things['things'][0]['thingName'].should.equal('51')
    things['things'][0]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/51')
    things['things'][-1]['thingName'].should.equal('100')
    things['things'][-1]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/100')

    things = client.list_things(nextToken=things['nextToken'])
    things.should.have.key('nextToken')
    things.should.have.key('things').which.should.have.length_of(50)
    things['things'][0]['thingName'].should.equal('101')
    things['things'][0]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/101')
    things['things'][-1]['thingName'].should.equal('150')
    things['things'][-1]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/150')

    things = client.list_things(nextToken=things['nextToken'])
    things.should_not.have.key('nextToken')
    things.should.have.key('things').which.should.have.length_of(50)
    things['things'][0]['thingName'].should.equal('151')
    things['things'][0]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/151')
    things['things'][-1]['thingName'].should.equal('200')
    things['things'][-1]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/200')


@mock_iot
def test_list_things_with_attribute_and_thing_type_filter_and_next_token():
    client = boto3.client('iot', region_name='ap-northeast-1')
    client.create_thing_type(thingTypeName='my-thing-type')

    for i in range(0, 200):
        if not (i + 1) % 3:
            attribute_payload = {
                'attributes': {
                    'foo': 'bar'
                }
            }
        elif not (i + 1) % 5:
            attribute_payload = {
                'attributes': {
                    'bar': 'foo'
                }
            }
        else:
            attribute_payload = {}

        if not (i + 1) % 2:
            thing_type_name = 'my-thing-type'
            client.create_thing(thingName=str(i + 1), thingTypeName=thing_type_name, attributePayload=attribute_payload)
        else:
            client.create_thing(thingName=str(i + 1), attributePayload=attribute_payload)

    # Test filter for thingTypeName
    things = client.list_things(thingTypeName=thing_type_name)
    things.should.have.key('nextToken')
    things.should.have.key('things').which.should.have.length_of(50)
    things['things'][0]['thingName'].should.equal('2')
    things['things'][0]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/2')
    things['things'][-1]['thingName'].should.equal('100')
    things['things'][-1]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/100')
    all(item['thingTypeName'] == thing_type_name for item in things['things'])

    things = client.list_things(nextToken=things['nextToken'], thingTypeName=thing_type_name)
    things.should_not.have.key('nextToken')
    things.should.have.key('things').which.should.have.length_of(50)
    things['things'][0]['thingName'].should.equal('102')
    things['things'][0]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/102')
    things['things'][-1]['thingName'].should.equal('200')
    things['things'][-1]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/200')
    all(item['thingTypeName'] == thing_type_name for item in things['things'])

    # Test filter for attributes
    things = client.list_things(attributeName='foo', attributeValue='bar')
    things.should.have.key('nextToken')
    things.should.have.key('things').which.should.have.length_of(50)
    things['things'][0]['thingName'].should.equal('3')
    things['things'][0]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/3')
    things['things'][-1]['thingName'].should.equal('150')
    things['things'][-1]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/150')
    all(item['attributes'] == {'foo': 'bar'} for item in things['things'])

    things = client.list_things(nextToken=things['nextToken'], attributeName='foo', attributeValue='bar')
    things.should_not.have.key('nextToken')
    things.should.have.key('things').which.should.have.length_of(16)
    things['things'][0]['thingName'].should.equal('153')
    things['things'][0]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/153')
    things['things'][-1]['thingName'].should.equal('198')
    things['things'][-1]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/198')
    all(item['attributes'] == {'foo': 'bar'} for item in things['things'])

    # Test filter for attributes and thingTypeName
    things = client.list_things(thingTypeName=thing_type_name, attributeName='foo', attributeValue='bar')
    things.should_not.have.key('nextToken')
    things.should.have.key('things').which.should.have.length_of(33)
    things['things'][0]['thingName'].should.equal('6')
    things['things'][0]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/6')
    things['things'][-1]['thingName'].should.equal('198')
    things['things'][-1]['thingArn'].should.equal('arn:aws:iot:ap-northeast-1:1:thing/198')
    all(item['attributes'] == {'foo': 'bar'} and item['thingTypeName'] == thing_type_name for item in things['things'])


@mock_iot
def test_certs():
    client = boto3.client('iot', region_name='us-east-1')
    cert = client.create_keys_and_certificate(setAsActive=True)
    cert.should.have.key('certificateArn').which.should_not.be.none
    cert.should.have.key('certificateId').which.should_not.be.none
    cert.should.have.key('certificatePem').which.should_not.be.none
    cert.should.have.key('keyPair')
    cert['keyPair'].should.have.key('PublicKey').which.should_not.be.none
    cert['keyPair'].should.have.key('PrivateKey').which.should_not.be.none
    cert_id = cert['certificateId']

    cert = client.describe_certificate(certificateId=cert_id)
    cert.should.have.key('certificateDescription')
    cert_desc = cert['certificateDescription']
    cert_desc.should.have.key('certificateArn').which.should_not.be.none
    cert_desc.should.have.key('certificateId').which.should_not.be.none
    cert_desc.should.have.key('certificatePem').which.should_not.be.none
    cert_desc.should.have.key('status').which.should.equal('ACTIVE')
    cert_pem = cert_desc['certificatePem']

    res = client.list_certificates()
    for cert in res['certificates']:
        cert.should.have.key('certificateArn').which.should_not.be.none
        cert.should.have.key('certificateId').which.should_not.be.none
        cert.should.have.key('status').which.should_not.be.none
        cert.should.have.key('creationDate').which.should_not.be.none

    client.update_certificate(certificateId=cert_id, newStatus='REVOKED')
    cert = client.describe_certificate(certificateId=cert_id)
    cert_desc = cert['certificateDescription']
    cert_desc.should.have.key('status').which.should.equal('REVOKED')

    client.delete_certificate(certificateId=cert_id)
    res = client.list_certificates()
    res.should.have.key('certificates')

    # Test register_certificate flow
    cert = client.register_certificate(certificatePem=cert_pem, setAsActive=True)
    cert.should.have.key('certificateId').which.should_not.be.none
    cert.should.have.key('certificateArn').which.should_not.be.none
    cert_id = cert['certificateId']

    res = client.list_certificates()
    res.should.have.key('certificates').which.should.have.length_of(1)
    for cert in res['certificates']:
        cert.should.have.key('certificateArn').which.should_not.be.none
        cert.should.have.key('certificateId').which.should_not.be.none
        cert.should.have.key('status').which.should_not.be.none
        cert.should.have.key('creationDate').which.should_not.be.none

    client.update_certificate(certificateId=cert_id, newStatus='REVOKED')
    cert = client.describe_certificate(certificateId=cert_id)
    cert_desc = cert['certificateDescription']
    cert_desc.should.have.key('status').which.should.equal('REVOKED')

    client.delete_certificate(certificateId=cert_id)
    res = client.list_certificates()
    res.should.have.key('certificates')


@mock_iot
def test_delete_policy_validation():
    doc = """{
    "Version": "2012-10-17",
    "Statement":[
        {
            "Effect":"Allow",
            "Action":[
                "iot: *"
            ],
            "Resource":"*"
        }
      ]
    }
    """
    client = boto3.client('iot', region_name='ap-northeast-1')
    cert = client.create_keys_and_certificate(setAsActive=True)
    cert_arn = cert['certificateArn']
    policy_name = 'my-policy'
    client.create_policy(policyName=policy_name, policyDocument=doc)
    client.attach_principal_policy(policyName=policy_name, principal=cert_arn)

    with assert_raises(ClientError) as e:
        client.delete_policy(policyName=policy_name)
    e.exception.response['Error']['Message'].should.contain(
        'The policy cannot be deleted as the policy is attached to one or more principals (name=%s)' % policy_name)
    res = client.list_policies()
    res.should.have.key('policies').which.should.have.length_of(1)

    client.detach_principal_policy(policyName=policy_name, principal=cert_arn)
    client.delete_policy(policyName=policy_name)
    res = client.list_policies()
    res.should.have.key('policies').which.should.have.length_of(0)


@mock_iot
def test_delete_certificate_validation():
    doc = """{
    "Version": "2012-10-17",
    "Statement":[
        {
            "Effect":"Allow",
            "Action":[
                "iot: *"
            ],
            "Resource":"*"
        }
      ]
    }
    """
    client = boto3.client('iot', region_name='ap-northeast-1')
    cert = client.create_keys_and_certificate(setAsActive=True)
    cert_id = cert['certificateId']
    cert_arn = cert['certificateArn']
    policy_name = 'my-policy'
    thing_name = 'thing-1'
    client.create_policy(policyName=policy_name, policyDocument=doc)
    client.attach_principal_policy(policyName=policy_name, principal=cert_arn)
    client.create_thing(thingName=thing_name)
    client.attach_thing_principal(thingName=thing_name, principal=cert_arn)

    with assert_raises(ClientError) as e:
        client.delete_certificate(certificateId=cert_id)
    e.exception.response['Error']['Message'].should.contain(
        'Certificate must be deactivated (not ACTIVE) before deletion.')
    res = client.list_certificates()
    res.should.have.key('certificates').which.should.have.length_of(1)

    client.update_certificate(certificateId=cert_id, newStatus='REVOKED')
    with assert_raises(ClientError) as e:
        client.delete_certificate(certificateId=cert_id)
    e.exception.response['Error']['Message'].should.contain(
        'Things must be detached before deletion (arn: %s)' % cert_arn)
    res = client.list_certificates()
    res.should.have.key('certificates').which.should.have.length_of(1)

    client.detach_thing_principal(thingName=thing_name, principal=cert_arn)
    with assert_raises(ClientError) as e:
        client.delete_certificate(certificateId=cert_id)
    e.exception.response['Error']['Message'].should.contain(
        'Certificate policies must be detached before deletion (arn: %s)' % cert_arn)
    res = client.list_certificates()
    res.should.have.key('certificates').which.should.have.length_of(1)

    client.detach_principal_policy(policyName=policy_name, principal=cert_arn)
    client.delete_certificate(certificateId=cert_id)
    res = client.list_certificates()
    res.should.have.key('certificates').which.should.have.length_of(0)


@mock_iot
def test_certs_create_inactive():
    client = boto3.client('iot', region_name='ap-northeast-1')
    cert = client.create_keys_and_certificate(setAsActive=False)
    cert_id = cert['certificateId']

    cert = client.describe_certificate(certificateId=cert_id)
    cert.should.have.key('certificateDescription')
    cert_desc = cert['certificateDescription']
    cert_desc.should.have.key('status').which.should.equal('INACTIVE')

    client.update_certificate(certificateId=cert_id, newStatus='ACTIVE')
    cert = client.describe_certificate(certificateId=cert_id)
    cert.should.have.key('certificateDescription')
    cert_desc = cert['certificateDescription']
    cert_desc.should.have.key('status').which.should.equal('ACTIVE')


@mock_iot
def test_policy():
    client = boto3.client('iot', region_name='ap-northeast-1')
    name = 'my-policy'
    doc = '{}'
    policy = client.create_policy(policyName=name, policyDocument=doc)
    policy.should.have.key('policyName').which.should.equal(name)
    policy.should.have.key('policyArn').which.should_not.be.none
    policy.should.have.key('policyDocument').which.should.equal(doc)
    policy.should.have.key('policyVersionId').which.should.equal('1')

    policy = client.get_policy(policyName=name)
    policy.should.have.key('policyName').which.should.equal(name)
    policy.should.have.key('policyArn').which.should_not.be.none
    policy.should.have.key('policyDocument').which.should.equal(doc)
    policy.should.have.key('defaultVersionId').which.should.equal('1')

    res = client.list_policies()
    res.should.have.key('policies').which.should.have.length_of(1)
    for policy in res['policies']:
        policy.should.have.key('policyName').which.should_not.be.none
        policy.should.have.key('policyArn').which.should_not.be.none

    client.delete_policy(policyName=name)
    res = client.list_policies()
    res.should.have.key('policies').which.should.have.length_of(0)


@mock_iot
def test_principal_policy():
    client = boto3.client('iot', region_name='ap-northeast-1')
    policy_name = 'my-policy'
    doc = '{}'
    client.create_policy(policyName=policy_name, policyDocument=doc)
    cert = client.create_keys_and_certificate(setAsActive=True)
    cert_arn = cert['certificateArn']

    client.attach_policy(policyName=policy_name, target=cert_arn)

    res = client.list_principal_policies(principal=cert_arn)
    res.should.have.key('policies').which.should.have.length_of(1)
    for policy in res['policies']:
        policy.should.have.key('policyName').which.should_not.be.none
        policy.should.have.key('policyArn').which.should_not.be.none

    # do nothing if policy have already attached to certificate
    client.attach_policy(policyName=policy_name, target=cert_arn)

    res = client.list_principal_policies(principal=cert_arn)
    res.should.have.key('policies').which.should.have.length_of(1)
    for policy in res['policies']:
        policy.should.have.key('policyName').which.should_not.be.none
        policy.should.have.key('policyArn').which.should_not.be.none

    res = client.list_policy_principals(policyName=policy_name)
    res.should.have.key('principals').which.should.have.length_of(1)
    for principal in res['principals']:
        principal.should_not.be.none

    client.detach_policy(policyName=policy_name, target=cert_arn)
    res = client.list_principal_policies(principal=cert_arn)
    res.should.have.key('policies').which.should.have.length_of(0)
    res = client.list_policy_principals(policyName=policy_name)
    res.should.have.key('principals').which.should.have.length_of(0)
    with assert_raises(ClientError) as e:
        client.detach_policy(policyName=policy_name, target=cert_arn)
    e.exception.response['Error']['Code'].should.equal('ResourceNotFoundException')


@mock_iot
def test_principal_policy_deprecated():
    client = boto3.client('iot', region_name='ap-northeast-1')
    policy_name = 'my-policy'
    doc = '{}'
    policy = client.create_policy(policyName=policy_name, policyDocument=doc)
    cert = client.create_keys_and_certificate(setAsActive=True)
    cert_arn = cert['certificateArn']

    client.attach_principal_policy(policyName=policy_name, principal=cert_arn)

    res = client.list_principal_policies(principal=cert_arn)
    res.should.have.key('policies').which.should.have.length_of(1)
    for policy in res['policies']:
        policy.should.have.key('policyName').which.should_not.be.none
        policy.should.have.key('policyArn').which.should_not.be.none

    res = client.list_policy_principals(policyName=policy_name)
    res.should.have.key('principals').which.should.have.length_of(1)
    for principal in res['principals']:
        principal.should_not.be.none

    client.detach_principal_policy(policyName=policy_name, principal=cert_arn)
    res = client.list_principal_policies(principal=cert_arn)
    res.should.have.key('policies').which.should.have.length_of(0)
    res = client.list_policy_principals(policyName=policy_name)
    res.should.have.key('principals').which.should.have.length_of(0)


@mock_iot
def test_principal_thing():
    client = boto3.client('iot', region_name='ap-northeast-1')
    thing_name = 'my-thing'
    thing = client.create_thing(thingName=thing_name)
    cert = client.create_keys_and_certificate(setAsActive=True)
    cert_arn = cert['certificateArn']

    client.attach_thing_principal(thingName=thing_name, principal=cert_arn)

    res = client.list_principal_things(principal=cert_arn)
    res.should.have.key('things').which.should.have.length_of(1)
    for thing in res['things']:
        thing.should_not.be.none
    res = client.list_thing_principals(thingName=thing_name)
    res.should.have.key('principals').which.should.have.length_of(1)
    for principal in res['principals']:
        principal.should_not.be.none

    client.detach_thing_principal(thingName=thing_name, principal=cert_arn)
    res = client.list_principal_things(principal=cert_arn)
    res.should.have.key('things').which.should.have.length_of(0)
    res = client.list_thing_principals(thingName=thing_name)
    res.should.have.key('principals').which.should.have.length_of(0)


@mock_iot
def test_thing_groups():
    client = boto3.client('iot', region_name='ap-northeast-1')
    group_name = 'my-group-name'

    # thing group
    thing_group = client.create_thing_group(thingGroupName=group_name)
    thing_group.should.have.key('thingGroupName').which.should.equal(group_name)
    thing_group.should.have.key('thingGroupArn')

    res = client.list_thing_groups()
    res.should.have.key('thingGroups').which.should.have.length_of(1)
    for thing_group in res['thingGroups']:
        thing_group.should.have.key('groupName').which.should_not.be.none
        thing_group.should.have.key('groupArn').which.should_not.be.none

    thing_group = client.describe_thing_group(thingGroupName=group_name)
    thing_group.should.have.key('thingGroupName').which.should.equal(group_name)
    thing_group.should.have.key('thingGroupProperties')
    thing_group.should.have.key('thingGroupMetadata')
    thing_group.should.have.key('version')

    # delete thing group
    client.delete_thing_group(thingGroupName=group_name)
    res = client.list_thing_groups()
    res.should.have.key('thingGroups').which.should.have.length_of(0)

    # props create test
    props = {
        'thingGroupDescription': 'my first thing group',
        'attributePayload': {
            'attributes': {
                'key1': 'val01',
                'Key02': 'VAL2'
            }
        }
    }
    thing_group = client.create_thing_group(thingGroupName=group_name, thingGroupProperties=props)
    thing_group.should.have.key('thingGroupName').which.should.equal(group_name)
    thing_group.should.have.key('thingGroupArn')

    thing_group = client.describe_thing_group(thingGroupName=group_name)
    thing_group.should.have.key('thingGroupProperties') \
        .which.should.have.key('attributePayload') \
        .which.should.have.key('attributes')
    res_props = thing_group['thingGroupProperties']['attributePayload']['attributes']
    res_props.should.have.key('key1').which.should.equal('val01')
    res_props.should.have.key('Key02').which.should.equal('VAL2')

    # props update test with merge
    new_props = {
        'attributePayload': {
            'attributes': {
                'k3': 'v3'
            },
            'merge': True
        }
    }
    client.update_thing_group(
        thingGroupName=group_name,
        thingGroupProperties=new_props
    )
    thing_group = client.describe_thing_group(thingGroupName=group_name)
    thing_group.should.have.key('thingGroupProperties') \
        .which.should.have.key('attributePayload') \
        .which.should.have.key('attributes')
    res_props = thing_group['thingGroupProperties']['attributePayload']['attributes']
    res_props.should.have.key('key1').which.should.equal('val01')
    res_props.should.have.key('Key02').which.should.equal('VAL2')

    res_props.should.have.key('k3').which.should.equal('v3')

    # props update test
    new_props = {
        'attributePayload': {
            'attributes': {
                'k4': 'v4'
            }
        }
    }
    client.update_thing_group(
        thingGroupName=group_name,
        thingGroupProperties=new_props
    )
    thing_group = client.describe_thing_group(thingGroupName=group_name)
    thing_group.should.have.key('thingGroupProperties') \
        .which.should.have.key('attributePayload') \
        .which.should.have.key('attributes')
    res_props = thing_group['thingGroupProperties']['attributePayload']['attributes']
    res_props.should.have.key('k4').which.should.equal('v4')
    res_props.should_not.have.key('key1')


@mock_iot
def test_thing_group_relations():
    client = boto3.client('iot', region_name='ap-northeast-1')
    name = 'my-thing'
    group_name = 'my-group-name'

    # thing group
    thing_group = client.create_thing_group(thingGroupName=group_name)
    thing_group.should.have.key('thingGroupName').which.should.equal(group_name)
    thing_group.should.have.key('thingGroupArn')

    # thing
    thing = client.create_thing(thingName=name)
    thing.should.have.key('thingName').which.should.equal(name)
    thing.should.have.key('thingArn')

    # add in 4 way
    client.add_thing_to_thing_group(
        thingGroupName=group_name,
        thingName=name
    )
    client.add_thing_to_thing_group(
        thingGroupArn=thing_group['thingGroupArn'],
        thingArn=thing['thingArn']
    )
    client.add_thing_to_thing_group(
        thingGroupName=group_name,
        thingArn=thing['thingArn']
    )
    client.add_thing_to_thing_group(
        thingGroupArn=thing_group['thingGroupArn'],
        thingName=name
    )

    things = client.list_things_in_thing_group(
        thingGroupName=group_name
    )
    things.should.have.key('things')
    things['things'].should.have.length_of(1)

    thing_groups = client.list_thing_groups_for_thing(
        thingName=name
    )
    thing_groups.should.have.key('thingGroups')
    thing_groups['thingGroups'].should.have.length_of(1)

    # remove in 4 way
    client.remove_thing_from_thing_group(
        thingGroupName=group_name,
        thingName=name
    )
    client.remove_thing_from_thing_group(
        thingGroupArn=thing_group['thingGroupArn'],
        thingArn=thing['thingArn']
    )
    client.remove_thing_from_thing_group(
        thingGroupName=group_name,
        thingArn=thing['thingArn']
    )
    client.remove_thing_from_thing_group(
        thingGroupArn=thing_group['thingGroupArn'],
        thingName=name
    )
    things = client.list_things_in_thing_group(
        thingGroupName=group_name
    )
    things.should.have.key('things')
    things['things'].should.have.length_of(0)

    # update thing group for thing
    client.update_thing_groups_for_thing(
        thingName=name,
        thingGroupsToAdd=[
            group_name
        ]
    )
    things = client.list_things_in_thing_group(
        thingGroupName=group_name
    )
    things.should.have.key('things')
    things['things'].should.have.length_of(1)

    client.update_thing_groups_for_thing(
        thingName=name,
        thingGroupsToRemove=[
            group_name
        ]
    )
    things = client.list_things_in_thing_group(
        thingGroupName=group_name
    )
    things.should.have.key('things')
    things['things'].should.have.length_of(0)


@mock_iot
def test_create_job():
    client = boto3.client('iot', region_name='eu-west-1')
    name = "my-thing"
    job_id = "TestJob"
    # thing
    thing = client.create_thing(thingName=name)
    thing.should.have.key('thingName').which.should.equal(name)
    thing.should.have.key('thingArn')

    # job document
    job_document = {
        "field": "value"
    }

    job = client.create_job(
        jobId=job_id,
        targets=[thing["thingArn"]],
        document=json.dumps(job_document),
        description="Description",
        presignedUrlConfig={
            'roleArn': 'arn:aws:iam::1:role/service-role/iot_job_role',
            'expiresInSec': 123
        },
        targetSelection="CONTINUOUS",
        jobExecutionsRolloutConfig={
            'maximumPerMinute': 10
        }
    )

    job.should.have.key('jobId').which.should.equal(job_id)
    job.should.have.key('jobArn')
    job.should.have.key('description')


@mock_iot
def test_describe_job():
    client = boto3.client('iot', region_name='eu-west-1')
    name = "my-thing"
    job_id = "TestJob"
    # thing
    thing = client.create_thing(thingName=name)
    thing.should.have.key('thingName').which.should.equal(name)
    thing.should.have.key('thingArn')

    job = client.create_job(
        jobId=job_id,
        targets=[thing["thingArn"]],
        documentSource="https://s3-eu-west-1.amazonaws.com/bucket-name/job_document.json",
        presignedUrlConfig={
            'roleArn': 'arn:aws:iam::1:role/service-role/iot_job_role',
            'expiresInSec': 123
        },
        targetSelection="CONTINUOUS",
        jobExecutionsRolloutConfig={
            'maximumPerMinute': 10
        }
    )

    job.should.have.key('jobId').which.should.equal(job_id)
    job.should.have.key('jobArn')

    job = client.describe_job(jobId=job_id)
    job.should.have.key('documentSource')
    job.should.have.key('job')
    job.should.have.key('job').which.should.have.key("jobArn")
    job.should.have.key('job').which.should.have.key("jobId").which.should.equal(job_id)
    job.should.have.key('job').which.should.have.key("targets")
    job.should.have.key('job').which.should.have.key("jobProcessDetails")
    job.should.have.key('job').which.should.have.key("lastUpdatedAt")
    job.should.have.key('job').which.should.have.key("createdAt")
    job.should.have.key('job').which.should.have.key("jobExecutionsRolloutConfig")
    job.should.have.key('job').which.should.have.key("targetSelection").which.should.equal("CONTINUOUS")
    job.should.have.key('job').which.should.have.key("presignedUrlConfig")
    job.should.have.key('job').which.should.have.key("presignedUrlConfig").which.should.have.key(
        "roleArn").which.should.equal('arn:aws:iam::1:role/service-role/iot_job_role')
    job.should.have.key('job').which.should.have.key("presignedUrlConfig").which.should.have.key(
        "expiresInSec").which.should.equal(123)
    job.should.have.key('job').which.should.have.key("jobExecutionsRolloutConfig").which.should.have.key(
        "maximumPerMinute").which.should.equal(10)


@mock_iot
def test_describe_job_1():
    client = boto3.client('iot', region_name='eu-west-1')
    name = "my-thing"
    job_id = "TestJob"
    # thing
    thing = client.create_thing(thingName=name)
    thing.should.have.key('thingName').which.should.equal(name)
    thing.should.have.key('thingArn')

    # job document
    job_document = {
        "field": "value"
    }

    job = client.create_job(
        jobId=job_id,
        targets=[thing["thingArn"]],
        document=json.dumps(job_document),
        presignedUrlConfig={
            'roleArn': 'arn:aws:iam::1:role/service-role/iot_job_role',
            'expiresInSec': 123
        },
        targetSelection="CONTINUOUS",
        jobExecutionsRolloutConfig={
            'maximumPerMinute': 10
        }
    )

    job.should.have.key('jobId').which.should.equal(job_id)
    job.should.have.key('jobArn')

    job = client.describe_job(jobId=job_id)
    job.should.have.key('job')
    job.should.have.key('job').which.should.have.key("jobArn")
    job.should.have.key('job').which.should.have.key("jobId").which.should.equal(job_id)
    job.should.have.key('job').which.should.have.key("targets")
    job.should.have.key('job').which.should.have.key("jobProcessDetails")
    job.should.have.key('job').which.should.have.key("lastUpdatedAt")
    job.should.have.key('job').which.should.have.key("createdAt")
    job.should.have.key('job').which.should.have.key("jobExecutionsRolloutConfig")
    job.should.have.key('job').which.should.have.key("targetSelection").which.should.equal("CONTINUOUS")
    job.should.have.key('job').which.should.have.key("presignedUrlConfig")
    job.should.have.key('job').which.should.have.key("presignedUrlConfig").which.should.have.key(
        "roleArn").which.should.equal('arn:aws:iam::1:role/service-role/iot_job_role')
    job.should.have.key('job').which.should.have.key("presignedUrlConfig").which.should.have.key(
        "expiresInSec").which.should.equal(123)
    job.should.have.key('job').which.should.have.key("jobExecutionsRolloutConfig").which.should.have.key(
        "maximumPerMinute").which.should.equal(10)
